# Copyright (c) Eric J. Wang. All Rights Reserved.
# ------------------------------------------------------------------------
# Modified from alpaca-lora/finetune.py (https://github.com/tloen/alpaca-lora/)
# Copyright (c) 2023 SenseTime. All Rights Reserved.
# ------------------------------------------------------------------------

import os
from typing import List
import subprocess

import fire
import torch
import transformers
from datasets import load_dataset, concatenate_datasets
import numpy as np
import torch.distributed as dist
from peft import LoraConfig, get_peft_model, set_peft_model_state_dict

from transformers import AutoTokenizer, AutoModel
import deepspeed

from utils.prompter import Prompter

def init_dist(port):
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    if "SLURM_JOB_ID" in os.environ:
        node_list = os.environ['SLURM_NODELIST']
        addr = subprocess.getoutput(f'scontrol show hostname {node_list} | head -n1')
        os.environ['MASTER_ADDR'] = addr
        os.environ['MASTER_PORT'] = str(port)
        print(addr, port)
        proc_id = int(os.environ['SLURM_PROCID'])
        ntasks = int(os.environ['SLURM_NTASKS'])
        rank = proc_id
        world_size = ntasks

        os.environ['WORLD_SIZE'] = str(ntasks)
        os.environ['RANK'] = str(proc_id)

        dist.init_process_group(backend='nccl')
        local_rank = int(os.environ['SLURM_LOCALID'])

        print("comm: ", world_size, rank, local_rank)
        os.environ['LOCAL_RANK'] = str(local_rank)
        torch.cuda.set_device(local_rank)
    else:
        local_rank = int(os.getenv("LOCAL_RANK", "0"))
        world_size = int(os.getenv("WORLD_SIZE", "1"))
        torch.cuda.set_device(local_rank)
        rank = os.getenv("RANK", None)
        if rank:
            rank = int(rank)
        else:
            rank = local_rank

    deepspeed.init_distributed()
    return rank

def train( # noqa C901
    # model/data params
    base_model: str = "",  # the only required argument
    data_path: str = "",
    output_dir: str = os.path.abspath("./output_dir"),
    # training hyperparams
    batch_size: int = 128,
    micro_batch_size: int = 4,
    num_epochs: int = 3,
    learning_rate: float = 3e-4,
    cutoff_len: int = 256,
    filter_nums: int = 512,
    val_set_size: int = 2000,
    # lora hyperparams
    lora_r: int = 8,
    lora_alpha: int = 16,
    lora_dropout: float = 0.05,
    lora_target_modules: List[str] = [
        "q_proj",
        "v_proj",
    ],
    modules_to_save: List[str] = None,
    # llm hyperparams
    train_on_inputs: bool = True,  # if False, masks out inputs in loss
    add_eos_token: bool = False,
    group_by_length: bool = False,  # faster, but produces an odd training loss curve
    # wandb params
    wandb_project: str = "",
    wandb_run_name: str = "",
    wandb_watch: str = "",  # options: false | gradients | all
    wandb_log_model: str = "",  # options: false | true
    resume_from_checkpoint: str = None,  # either training checkpoint or final adapter
    prompt_template_name: str = "alpaca",  # The prompt template to use, will default to alpaca.
    deepspeed: str = None,  # deepspeed config file path.
    show_prompt: bool = False,
    nccl_port: int = 19080,
):
    rank = init_dist(nccl_port)
    if int(os.environ.get("RANK", 0)) == 0:
        print(
            f"Training Alpaca-LoRA model with params:\n"
            f"base_model: {base_model}\n"
            f"data_path: {data_path}\n"
            f"output_dir: {output_dir}\n"
            f"batch_size: {batch_size}\n"
            f"micro_batch_size: {micro_batch_size}\n"
            f"num_epochs: {num_epochs}\n"
            f"learning_rate: {learning_rate}\n"
            f"cutoff_len: {cutoff_len}\n"
            f"filter_nums: {filter_nums}\n"
            f"val_set_size: {val_set_size}\n"
            f"lora_r: {lora_r}\n"
            f"lora_alpha: {lora_alpha}\n"
            f"lora_dropout: {lora_dropout}\n"
            f"lora_target_modules: {lora_target_modules}\n"
            f"modules_to_save: {modules_to_save}\n"
            f"train_on_inputs: {train_on_inputs}\n"
            f"add_eos_token: {add_eos_token}\n"
            f"group_by_length: {group_by_length}\n"
            f"wandb_project: {wandb_project}\n"
            f"wandb_run_name: {wandb_run_name}\n"
            f"wandb_watch: {wandb_watch}\n"
            f"wandb_log_model: {wandb_log_model}\n"
            f"resume_from_checkpoint: {resume_from_checkpoint or False}\n"
            f"prompt template: {prompt_template_name}\n"
            f"show prompt: {show_prompt}\n"
        )
    assert (
        base_model
    ), "Please specify a --base_model, e.g. --base_model='huggyllama/llama-7b'"

    gradient_accumulation_steps = batch_size // micro_batch_size
    prompter = Prompter.from_template(prompt_template_name, show=show_prompt)

    world_size = int(os.environ.get("WORLD_SIZE", 1))
    ddp = world_size != 1
    if ddp:
        gradient_accumulation_steps = gradient_accumulation_steps // world_size
    if rank == 0:
        print("gradient_accumulation_steps: ", gradient_accumulation_steps)

    # Only overwrite environ if wandb param passed
    if len(wandb_project) > 0:
        os.environ["WANDB_PROJECT"] = wandb_project
    if len(wandb_watch) > 0:
        os.environ["WANDB_WATCH"] = wandb_watch
    if len(wandb_log_model) > 0:
        os.environ["WANDB_LOG_MODEL"] = wandb_log_model
    os.environ['WANDB_DISABLED'] = 'true'

    model = AutoModel.from_pretrained(
        base_model,
        load_in_8bit=False,
        torch_dtype=torch.float16,
        trust_remote_code=True,
    )

    tokenizer = AutoTokenizer.from_pretrained(base_model, trust_remote_code=True)

    def tokenize(prompt, add_eos_token=True):
        # there's probably a way to do this with the tokenizer settings
        # but again, gotta move fast
        result = tokenizer(
            prompt,
            truncation=True,
            max_length=cutoff_len,
            padding=False,
            return_tensors=None,
        )
        if len(result["input_ids"]) > cutoff_len:
            return None
        if (result["input_ids"][-1] != tokenizer.eos_token_id and len(result["input_ids"]) < cutoff_len and add_eos_token
        ):
            result["input_ids"].append(tokenizer.eos_token_id)
            result["attention_mask"].append(1)

        result["labels"] = result["input_ids"].copy()

        return result

    def generate_and_tokenize_prompt(data_point):
        output = data_point.pop("output", None)
        full_prompt = prompter.generate_prompt(data_point, label=output)
        tokenized_full_prompt = tokenize(full_prompt)
        if not tokenized_full_prompt:
            return None
        train_on_inputs = True
        if not train_on_inputs:
            user_prompt = prompter.generate_prompt(data_point)
            tokenized_user_prompt = tokenize(
                user_prompt, add_eos_token=add_eos_token
            )
            user_prompt_len = len(tokenized_user_prompt["input_ids"])

            if add_eos_token:
                user_prompt_len -= 1

            tokenized_full_prompt["labels"] = [
                -100
            ] * user_prompt_len + tokenized_full_prompt["labels"][
                user_prompt_len:
            ]  # could be sped up, probably
        return tokenized_full_prompt

    modules_to_save = list(modules_to_save) if modules_to_save else None
    config = LoraConfig(
        r=lora_r,
        lora_alpha=lora_alpha,
        target_modules=lora_target_modules,
        modules_to_save=modules_to_save,
        lora_dropout=lora_dropout,
        bias="none",
        task_type="CAUSAL_LM",
    )
    model = get_peft_model(model, config)

    datas = []
    data_names = ''
    if os.path.isfile(data_path):
        if data_path.endswith(".json") or data_path.endswith(".jsonl"):
            datas.append(load_dataset("json", data_files=data_path, split='train'))
        else:
            datas.append(load_dataset(data_path))
    elif os.path.isdir(data_path):
        for root, dirs, files in os.walk(data_path):
            for file in files:
                if file.endswith(".json") or file.endswith(".jsonl"):
                    file_path = os.path.join(root, file)
                    aa = load_dataset("json", data_files=file_path, split='train')
                    key_list = aa.column_names
                    for key in key_list:
                        if key not in ['instruction', 'output', 'input']:
                            aa = aa.remove_columns(key)
                    datas.append(aa)
                    data_names += file + '\n'
    assert len(datas) > 0, "Invalid file"
    if data_names:
        print("Merge Dataset: \n", data_names)

    data = concatenate_datasets(datas)

    if resume_from_checkpoint:
        # Check the available weights and load them
        checkpoint_name = os.path.join(
            resume_from_checkpoint, "pytorch_model.bin"
        )  # Full checkpoint
        if not os.path.exists(checkpoint_name):
            checkpoint_name = os.path.join(
                resume_from_checkpoint, "adapter_model.bin"
            )  # only LoRA model - LoRA config above has to fit
            resume_from_checkpoint = (
                False  # So the trainer won't try loading its state
            )
        # The two files above have a different name depending on how they were saved, but are actually the same.
        if os.path.exists(checkpoint_name):
            print(f"Restarting from {checkpoint_name}")
            adapters_weights = torch.load(checkpoint_name)
            set_peft_model_state_dict(model, adapters_weights)
        else:
            print(f"Checkpoint {checkpoint_name} not found")

    if rank == 0:
        model.print_trainable_parameters()  # Be more transparent about the % of trainable params.

    if val_set_size > 0:
        train_val = data.train_test_split(
            test_size=val_set_size, shuffle=True, seed=42
        )
        train_data = (
            train_val["train"].shuffle().map(generate_and_tokenize_prompt, num_proc=16)
            .filter(lambda x: len(x["input_ids"]) < filter_nums)
        )
        val_data = (
            train_val["test"].shuffle().map(generate_and_tokenize_prompt, num_proc=16)
            .filter(lambda x: len(x["input_ids"]) < filter_nums)
        )
    else:
        train_data = data.shuffle().map(generate_and_tokenize_prompt, num_proc=16)\
            .filter(lambda x: len(x["input_ids"]) < filter_nums)
        val_data = None

    tlist = np.array([len(x) for x in train_data['input_ids']])

    if int(os.environ.get("RANK", 0)) == 0:
        print("\n\nTotal item: ", len(tlist))
        print("cut-off: ", cutoff_len)
        print("max-len token: ", np.max(tlist))
        print("min-len token: ", np.min(tlist))
        for target in [512, 1024, 2048, 4096]:
            count = np.sum(tlist > target)
            print(f"more than {target}-len: {count}")
        print(" \n")

    if not ddp and torch.cuda.device_count() > 1:
        # keeps Trainer from trying its own DataParallelism when more than 1 gpu is available
        model.is_parallelizable = True
        model.model_parallel = True

    trainer = transformers.Trainer(
        model=model,
        train_dataset=train_data,
        eval_dataset=val_data,
        args=transformers.TrainingArguments(
            per_device_train_batch_size=micro_batch_size,
            gradient_accumulation_steps=gradient_accumulation_steps,
            warmup_steps=100,
            num_train_epochs=num_epochs,
            learning_rate=learning_rate,
            fp16=True,
            logging_steps=10,
            optim="adamw_torch",
            eval_strategy="steps" if val_set_size > 0 else "no",
            save_strategy="steps",
            eval_steps=1000 if val_set_size > 0 else None,
            save_steps=20000000,
            output_dir=output_dir,
            save_total_limit=3,
            load_best_model_at_end=True if val_set_size > 0 else False,
            ddp_find_unused_parameters=False if ddp else None,
            group_by_length=group_by_length,
            report_to="tensorboard",
            run_name="testhahha",
            lr_scheduler_type='cosine',
            deepspeed=deepspeed,
            local_rank=rank,
        ),
        data_collator=transformers.DataCollatorForSeq2Seq(
            tokenizer, pad_to_multiple_of=8, return_tensors="pt", padding=True
        ),
    )
    model.config.use_cache = False

    trainer.train(resume_from_checkpoint=resume_from_checkpoint)

    model.save_pretrained(output_dir)

    if rank == 0:
        print(
            "\n If there's a warning about missing keys above, please disregard :)"
        )

if __name__ == "__main__":
    fire.Fire(train)



"""
The overall process of SimpleDirectoryReader is borrowed from LLAMA_INDEX, but we have added a customized part
based on it, that is, allowing users to register custom rules instead of processing only based on file suffixes.
"""
import os
import mimetypes
import multiprocessing
import fnmatch
from tqdm import tqdm
from datetime import datetime
from functools import reduce
from itertools import repeat
from typing import Dict, Optional, List, Callable, Type
from pathlib import Path, PurePosixPath, PurePath
from fsspec import AbstractFileSystem
from lazyllm import ModuleBase, LOG
from .doc_node import DocNode
from .readers import (ReaderBase, PDFReader, DocxReader, HWPReader, PPTXReader, ImageReader, IPYNBReader,
                      EpubReader, MarkdownReader, MboxReader, PandasCSVReader, PandasExcelReader, VideoAudioReader,
                      get_default_fs, is_default_fs)
from .global_metadata import (RAG_DOC_PATH, RAG_DOC_FILE_NAME, RAG_DOC_FILE_TYPE, RAG_DOC_FILE_SIZE,
                              RAG_DOC_CREATION_DATE, RAG_DOC_LAST_MODIFIED_DATE, RAG_DOC_LAST_ACCESSED_DATE)

def _file_timestamp_format(timestamp: float, include_time: bool = False) -> Optional[str]:
    try:
        if include_time:
            return datetime.utcfromtimestamp(timestamp).strftime("%Y-%m-%dT%H:%M:%SZ")
        return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d")
    except Exception:
        return None

class _DefaultFileMetadataFunc:
    def __init__(self, fs: Optional[AbstractFileSystem] = None):
        self._fs = fs or get_default_fs()

    def __call__(self, file_path: str) -> Dict:
        stat_result = self._fs.stat(file_path)

        try:
            file_name = os.path.basename(str(stat_result['name']))
        except Exception:
            file_name = os.path.basename(file_path)

        creation_date = _file_timestamp_format(stat_result.get("created"))
        last_modified_date = _file_timestamp_format(stat_result.get("mtime"))
        last_accessed_date = _file_timestamp_format(stat_result.get("atime"))
        default_meta = {
            RAG_DOC_PATH: file_path,
            RAG_DOC_FILE_NAME: file_name,
            RAG_DOC_FILE_TYPE: mimetypes.guess_type(file_path)[0],
            RAG_DOC_FILE_SIZE: stat_result.get("size"),
            RAG_DOC_CREATION_DATE: creation_date,
            RAG_DOC_LAST_MODIFIED_DATE: last_modified_date,
            RAG_DOC_LAST_ACCESSED_DATE: last_accessed_date,
        }

        return {meta_key: meta_value for meta_key, meta_value in default_meta.items() if meta_value is not None}

class SimpleDirectoryReader(ModuleBase):
    default_file_readers: Dict[str, Type[ReaderBase]] = {
        "*.pdf": PDFReader,
        "*.docx": DocxReader,
        "*.hwp": HWPReader,
        "*.pptx": PPTXReader,
        "*.ppt": PPTXReader,
        "*.pptm": PPTXReader,
        "*.gif": ImageReader,
        "*.jpeg": ImageReader,
        "*.jpg": ImageReader,
        "*.png": ImageReader,
        "*.webp": ImageReader,
        "*.ipynb": IPYNBReader,
        "*.epub": EpubReader,
        "*.md": MarkdownReader,
        "*.mbox": MboxReader,
        "*.csv": PandasCSVReader,
        "*.xls": PandasExcelReader,
        "*.xlsx": PandasExcelReader,
        "*.mp3": VideoAudioReader,
        "*.mp4": VideoAudioReader,
    }

    def __init__(self, input_dir: Optional[str] = None, input_files: Optional[List] = None,
                 exclude: Optional[List] = None, exclude_hidden: bool = True, recursive: bool = False,
                 encoding: str = "utf-8", filename_as_id: bool = False, required_exts: Optional[List[str]] = None,
                 file_extractor: Optional[Dict[str, Callable]] = None, fs: Optional[AbstractFileSystem] = None,
                 metadata_genf: Optional[Callable[[str], Dict]] = None, num_files_limit: Optional[int] = None,
                 return_trace: bool = False, metadatas: Optional[Dict] = None) -> None:
        super().__init__(return_trace=return_trace)

        if (not input_dir and not input_files) or (input_dir and input_files):
            raise ValueError("Must provide either `input_dir` or `input_files`.")

        self._fs = fs or get_default_fs()
        self._encoding = encoding

        self._exclude = exclude
        self._recursive = recursive
        self._exclude_hidden = exclude_hidden
        self._required_exts = required_exts
        self._num_files_limit = num_files_limit
        self._Path = Path if is_default_fs(self._fs) else PurePosixPath
        self._metadatas = metadatas

        if input_files:
            self._input_files = []
            for path in input_files:
                if not self._fs.isfile(path):
                    raise ValueError(f"File {path} does not exist.")
                input_file = self._Path(path)
                self._input_files.append(input_file)
        elif input_dir:
            if not self._fs.isdir(input_dir):
                raise ValueError(f"Directory {input_dir} does not exist.")
            self._input_dir = self._Path(input_dir)
            self._input_files = self._add_files(self._input_dir)

        self._file_extractor = file_extractor or {}

        self._metadata_genf = metadata_genf or _DefaultFileMetadataFunc(self._fs)
        self._filename_as_id = filename_as_id

    def _add_files(self, input_dir: Path) -> List[Path]:  # noqa: C901
        all_files = set()
        rejected_files = set()
        rejected_dirs = set()

        if self._exclude is not None:
            for excluded_pattern in self._exclude:
                if self._recursive:
                    excluded_glob = self._Path(input_dir) / self._Path("**") / excluded_pattern
                else:
                    excluded_glob = self._Path(input_dir) / excluded_pattern
                for file in self._fs.glob(str(excluded_glob)):
                    if self._fs.isdir(file):
                        rejected_dirs.add(self._Path(file))
                    else:
                        rejected_files.add(self._Path(file))

        file_refs: List[str] = []
        if self._recursive:
            file_refs = self._fs.glob(str(input_dir) + "/**/*")
        else:
            file_refs = self._fs.glob(str(input_dir) + "/*")

        for ref in file_refs:
            ref = self._Path(ref)
            is_dir = self._fs.isdir(ref)
            skip_hidden = self._exclude_hidden and self._is_hidden(ref)
            skip_bad_exts = (self._required_exts is not None and ref.suffix not in self._required_exts)
            skip_excluded = ref in rejected_files
            if not skip_excluded:
                if is_dir:
                    ref_parent_dir = ref
                else:
                    ref_parent_dir = self._fs._parent(ref)
                for rejected_dir in rejected_dirs:
                    if str(ref_parent_dir).startswith(str(rejected_dir)):
                        skip_excluded = True
                        LOG.warning(f"Skipping {ref} because it in parent dir "
                                    f"{ref_parent_dir} which is in {rejected_dir}.")
                        break

            if is_dir or skip_hidden or skip_bad_exts or skip_excluded:
                continue
            else:
                all_files.add(ref)

        new_input_files = sorted(all_files)

        if len(new_input_files) == 0:
            raise ValueError(f"No files found in {input_dir}.")
        if self._num_files_limit is not None and self._num_files_limit > 0:
            new_input_files = new_input_files[0: self._num_files_limit]

        LOG.debug(f"[SimpleDirectoryReader] Total files add: {len(new_input_files)}")

        LOG.info(f"input_files: {new_input_files}")
        return new_input_files

    def _is_hidden(self, path: Path) -> bool:
        return any(part.startswith(".") and part not in [".", ".."] for part in path.parts)

    def _exclude_metadata(self, documents: List[DocNode]) -> List[DocNode]:
        for doc in documents:
            doc._excluded_embed_metadata_keys.extend(
                ["file_name", "file_type", "file_size", "creation_date",
                 "last_modified_date", "last_accessed_date"])
            doc._excluded_llm_metadata_keys.extend(
                ["file_name", "file_type", "file_size", "creation_date",
                 "last_modified_date", "last_accessed_date"])
        return documents

    @staticmethod
    def load_file(input_file: Path, metadata_genf: Callable[[str], Dict], file_extractor: Dict[str, Callable],
                  filename_as_id: bool = False, encoding: str = "utf-8", pathm: PurePath = Path,
                  fs: Optional[AbstractFileSystem] = None, metadata: Optional[Dict] = None) -> List[DocNode]:
        metadata: dict = metadata or {}
        documents: List[DocNode] = []

        if metadata_genf is not None: metadata.update(metadata_genf(str(input_file)))
        file_reader_patterns = list(file_extractor.keys())

        for pattern in file_reader_patterns:
            pt = str(pathm(pattern))
            match_pattern = pt if pt.startswith("*") else os.path.join(str(pathm.cwd()), pt)
            if fnmatch.fnmatch(input_file, match_pattern):
                reader = file_extractor[pattern]
                reader = reader() if isinstance(reader, type) else reader
                kwargs = {"extra_info": metadata}
                if fs and not is_default_fs(fs): kwargs['fs'] = fs
                docs = reader(input_file, **kwargs)

                if filename_as_id:
                    for i, doc in enumerate(docs):
                        doc._uid = f"{input_file!s}_index_{i}"
                        doc.docpath = str(input_file)
                documents.extend(docs)
                break
        else:
            fs = fs or get_default_fs()
            with fs.open(input_file, encoding=encoding) as f:
                data = f.read().decode(encoding)

            doc = DocNode(text=data, global_metadata=metadata or {})
            doc.docpath = str(input_file)
            if filename_as_id: doc._uid = str(input_file)
            documents.append(doc)

        return documents

    def _load_data(self, show_progress: bool = False, num_workers: Optional[int] = None,
                   fs: Optional[AbstractFileSystem] = None) -> List[DocNode]:
        documents = []

        fs = fs or self._fs
        process_file = self._input_files
        file_readers = self._file_extractor.copy()
        for key, func in self.default_file_readers.items():
            if key not in file_readers: file_readers[key] = func

        if num_workers and num_workers >= 1:
            if num_workers > multiprocessing.cpu_count():
                LOG.warning("Specified num_workers exceed number of CPUs in the system. "
                            "Setting `num_workers` down to the maximum CPU count.")
            with multiprocessing.get_context("spawn").Pool(num_workers) as p:
                results = p.starmap(SimpleDirectoryReader.load_file,
                                    zip(process_file, repeat(self._metadata_genf), repeat(file_readers),
                                        repeat(self._filename_as_id), repeat(self._encoding), repeat(self._Path),
                                        repeat(self._fs), self._metadatas or repeat(None)))
                documents = reduce(lambda x, y: x + y, results)
        else:
            if show_progress:
                process_file = tqdm(self._input_files, desc="Loading files", unit="file")
            for input_file in process_file:
                documents.extend(
                    SimpleDirectoryReader.load_file(input_file=input_file, metadata_genf=self._metadata_genf,
                                                    file_extractor=file_readers, filename_as_id=self._filename_as_id,
                                                    encoding=self._encoding, pathm=self._Path, fs=self._fs))

        return self._exclude_metadata(documents)

    def forward(self, *args, **kwargs) -> List[DocNode]:
        return self._load_data(*args, **kwargs)



# -*- coding: utf-8 -*-

from .configs import config
from .configs import * # noqa F401 of Config
from .common import *  # noqa F403
from .launcher import LazyLLMLaunchersBase
from .flow import *  # noqa F403
from .components import (LazyLLMDataprocBase, LazyLLMFinetuneBase, LazyLLMDeployBase,
                         LazyLLMValidateBase, register as component_register, Prompter,
                         AlpacaPrompter, ChatPrompter, FastapiApp, JsonFormatter, FileFormatter)

from .module import (ModuleBase, ModuleBase as Module, UrlModule, TrainableModule, ActionModule,
                     ServerModule, TrialModule, register as module_register,
                     OnlineChatModule, OnlineEmbeddingModule, AutoModel)
from .client import redis_client
from .hook import LazyLLMHook
from .tools import (Document, Reranker, Retriever, WebModule, ToolManager, FunctionCall,
                    FunctionCallAgent, fc_register, ReactAgent, PlanAndSolveAgent, ReWOOAgent, SentenceSplitter,
                    LLMParser)
from .docs import add_doc

config.done()


del LazyLLMRegisterMetaClass  # noqa F821
del _get_base_cls_from_registry  # noqa F821


__all__ = [
    # components
    'LazyLLMDataprocBase',  #
    'LazyLLMFinetuneBase',        # finetune
    'LazyLLMDeployBase',          # deploy
    'LazyLLMValidateBase',        #
    'component_register',
    'Prompter',
    'AlpacaPrompter',
    'ChatPrompter',
    'FastapiApp',
    'JsonFormatter',
    'FileFormatter',

    # launcher
    'LazyLLMLaunchersBase',        # empty, slurm, sco

    # configs
    'Mode',

    # module
    'ModuleBase',
    'Module',
    'UrlModule',
    'TrainableModule',
    'ActionModule',
    'ServerModule',
    'WebModule',
    'TrialModule',
    'module_register',
    'OnlineChatModule',
    'OnlineEmbeddingModule',
    'AutoModel',

    # client
    'redis_client',

    # hook
    'LazyLLMHook',

    # tools
    'Document',
    'Retriever',
    'Reranker',
    'ToolManager',
    'FunctionCall',
    'FunctionCallAgent',
    'fc_register',
    "LLMParser",
    'ReactAgent',
    'PlanAndSolveAgent',
    'ReWOOAgent',
    'SentenceSplitter',

    # docs
    'add_doc',
]

__all__ += common.__all__  # noqa F405
__all__ += flow.__all__  # noqa F405



# flake8: noqa: E501,

model_groups = {
    "meta-llama-3": {
        "prompt_keys": {'sos': '<|start_header_id|>system<|end_header_id|>\n\n', 'soh': '<|start_header_id|>user<|end_header_id|>\n\n', 'soa': '<|start_header_id|>assistant<|end_header_id|>\n\n', 'eos': '<|eot_id|>', 'eoh': '<|eot_id|>', 'eoa': '<|eot_id|>', 'stop_words': ['<|eot_id|>', '<|end_of_text|>']},
    },
    "llama-2": {
        "prompt_keys": {
            'sos': '[INST] <<SYS>>\n', 'soa': ' [/INST] ', 'eos': '\n<</SYS>>\n\n', 'eoa': '</s>', 'separator': '<s>[INST] ',
            'system': ("\nYou are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe. Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. Please ensure that your responses are socially unbiased and positive in nature.\n\n"
                    "If a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct. If you don't know the answer to a question, please don't share false information.")
            },
    },
    "qwen": {
        "prompt_keys": {
            'sos': '<|im_start|>system\n', 'soh': '<|im_start|>user\n', 'soa': '<|im_start|>assistant\n', 'soe': '<|im_start|>environment\n', 'eos': '<|im_end|>\n', 'eoh': '<|im_end|>\n', 'eoa': '<|im_end|>', 'eoe': '<|im_end|>\n', 'separator': '\n', 'stop_words': ['<|im_end|>'], 'tool_start_token': '<|action_start|><|function|>', 'tool_end_token': '<|action_end|>', 'tool_args_token': "<|args|>",
            'system': "You are a helpful assistant."
            },
    },
    "deepseek-r1-distill-qwen": {
        "prompt_keys": {
            'sos': '<|im_start|>system\n', 'soh': '<|im_start|>user\n', 'soa': '<|im_start|>assistant\n\n<think>', 'soe': '<|im_start|>environment\n', 'eos': '<|im_end|>\n', 'eoh': '<|im_end|>\n', 'eoa': '<|im_end|>', 'eoe': '<|im_end|>\n', 'separator': '\n', 'stop_words': ['<|im_end|>'], 'tool_start_token': '<|action_start|><|function|>', 'tool_end_token': '<|action_end|>', 'tool_args_token': "<|args|>",
            'system': "You are a helpful assistant."
            },
    },
    "deepseek-r1-distill-llama": {
        "prompt_keys": {
            'sos': '<|start_header_id|>system<|end_header_id|>\n\n', 'soh': '<|start_header_id|>user<|end_header_id|>\n\n', 'soa': '<|start_header_id|>assistant<|end_header_id|>\n\n<think>', 'eos': '<|eot_id|>', 'eoh': '<|eot_id|>', 'eoa': '<|eot_id|>', 'stop_words': ['<|eot_id|>', '<|end_of_text|>', '<|eom_id|>'],
            'system': "You are a helpful assistant.",
            },
    },
    "deepseek": {
        "prompt_keys": {
            'sos': '<｜begin▁of▁sentence｜>', 'soh': '<｜User｜>', 'soa': '<｜Assistant｜>', 'eos': '', 'eoh': '', 'eoa': '<｜end▁of▁sentence｜>', 'stop_words': ['<｜end▁of▁sentence｜>'],
            'system': "",
            },
    },    
    "internlm": {
        "prompt_keys": {
            'sos': '<|System|>:', 'soh': '<|User|>:', 'soa': '<|Bot|>:', 'eos': '\n', 'eoh': '\n', 'eoa': '<eoa>', 'separator': '\n', 'stop_words': ['<eoa>'],
            'system': "You are an AI assistant whose name is InternLM (书生·浦语).\n- InternLM (书生·浦语) is a conversational language model that is developed by Shanghai AI Laboratory (上海人工智能实验室). It is designed to be helpful, honest, and harmless.\n- InternLM (书生·浦语) can understand and communicate fluently in the language chosen by the user such as English and 中文."
            },
    },
    "internlm2": {
        "prompt_keys": {
            'sos': '<|im_start|>system\n', 'soh': '<|im_start|>user\n', 'soa': '<|im_start|>assistant\n', 'soe': '<|im_start|>environment\n', 'plugin': '<|plugin|>', 'interpreter': '<|interpreter|>', 'eos': '<|im_end|>\n', 'eoh': '<|im_end|>\n', 'eoa': '<|im_end|>', 'eoe': '<|im_end|>\n', 'separator': '\n', 'stop_words': ['<|im_end|>'], 'tool_start_token': '<|function|>', 'tool_end_token': '<|action_end|>', 'tool_args_token': '<|args|>',
            'system': "You are an AI assistant whose name is InternLM (书生·浦语).\n- InternLM (书生·浦语) is a conversational language model that is developed by Shanghai AI Laboratory (上海人工智能实验室). It is designed to be helpful, honest, and harmless.\n- InternLM (书生·浦语) can understand and communicate fluently in the language chosen by the user such as English and 中文."
            },
    },
    "sensechat": {
        "prompt_keys": {
            'sos': '<|im_start|>system\n', 'soh': '<|im_start|>user\n', 'soa': '<|im_start|>assistant\n', 'soe': '<|im_start|>environment\n', 'plugin': '<|plugin|>', 'interpreter': '<|interpreter|>', 'eos': '<|im_end|>\n', 'eoh': '<|im_end|>\n', 'eoa': '<|im_end|>', 'eoe': '<|im_end|>\n', 'separator': '\n', 'stop_words': ['<|im_end|>'], 'tool_start_token': '<|function|>', 'tool_end_token': '<|action_end|>', 'tool_args_token': '<|args|>',
            'system': "You are an AI assistant whose name is SenseChat.\n- SenseChat is a conversational language model that is developed by SenseTime. It is designed to be helpful, honest, and harmless.\n- SenseChat can understand and communicate fluently in the language chosen by the user such as English and 中文."
            },
    },
    "internlm-xcomposer2": {
        "prompt_keys": {
            'sos': '[UNUSED_TOKEN_146]system\n', 'soh': '[UNUSED_TOKEN_146]user\n', 'soa': '[UNUSED_TOKEN_146]assistant\n', 'eos': '[UNUSED_TOKEN_145]\n', 'eoh': '[UNUSED_TOKEN_145]\n', 'eoa': '[UNUSED_TOKEN_145]\n', 'separator': '\n', 'stop_words': ['[UNUSED_TOKEN_145]'],
            'system': ("You are an AI assistant whose name is InternLM-XComposer (浦语·灵笔).\n- InternLM-XComposer (浦语·灵笔) is a multi-modality conversational language model that is developed by Shanghai AI Laboratory (上海人工智能实验室). It is designed to be helpful, honest, and harmless.\n"
                    "- InternLM-XComposer (浦语·灵笔) can understand and communicate fluently in the language chosen by the user such as English and 中文.\n- InternLM-XComposer (浦语·灵笔) is capable of comprehending and articulating responses effectively based on the provided image.")
            },
    },
    "internlm-xcomposer": {
        "prompt_keys": {},
    },
    "codellama": {
        "prompt_keys": {},
    },
    "baichuan": {
        "prompt_keys": {},
    },
    "baichuan2": {
        "prompt_keys": {'soh': '<reserved_106>', 'soa': '<reserved_107>'},
    },
    "chatglm3": {
        "prompt_keys": {'sos': '<|system|>\n', 'soh': '<|user|>\n', 'soa': '<|assistant|>\n', 'plugin': '<|observation|>\n', 'stop_words': ['<|user|>', '<|observation|>']},
    },
    "glm-4": {
        "prompt_keys": {'sos': '<|system|>\n', 'soh': '<|user|>\n', 'soa': '<|assistant|>\n', 'soe': '<|observation|>\n', 'eoe': '\n', 'plugin': '<|observation|>\n', 'stop_words': ['<|user|>', '<|observation|>'], 'tool_start_token': '✿FUNCTION✿: ', 'tool_args_token': '✿ARGS✿: ', 'tool_end_token': '\n'},
    },
}

model_provider = {
    "baichuan": {           
        "huggingface": "baichuan-inc",
        "modelscope": "baichuan-inc"
    },
    "chatglm": {
        "huggingface": "THUDM",
        "modelscope": "ZhipuAI"
    },
     "glm": {
        "huggingface": "THUDM",
        "modelscope": "ZhipuAI"
    },
    "codellama": {
        "huggingface": "meta-llama",
        "modelscope": "AI-ModelScope"
    },
    "llama": {
        "huggingface": "meta-llama",
        "modelscope": "modelscope"
    },
    "meta-llama": {
        "huggingface": "meta-llama",
        "modelscope": "LLM-Research"
    },
    "internlm": {
        "huggingface": "internlm",
        "modelscope": "Shanghai_AI_Laboratory"
    },
    "sensechat": {
        "huggingface": "sensetime",
        "modelscope": "sensetime"
    },
    "internvl": {
        "huggingface": "OpenGVLab",
        "modelscope": "OpenGVLab"
    },
    "qwen": {
        "huggingface": "Qwen",
        "modelscope": "qwen"
    },
    "deepseek":{
        "huggingface": "deepseek-ai",
        "modelscope": "deepseek-ai"
    }
}

model_name_mapping = {
    "bark": {
        "source": {
            "huggingface": "suno/bark",
            "modelscope": "mapjack/bark"
        },
        "type": "tts"
    },
    "bge-large-zh-v1.5": {
        "source": {
            "huggingface": "BAAI/bge-large-zh-v1.5",
            "modelscope": "AI-ModelScope/bge-large-zh-v1.5"
        },
        "type": "embed"
    },
    "bge-m3": {
        "source": {
            "huggingface": "BAAI/bge-m3",
            "modelscope": "Xorbits/bge-m3"
        },
        "type": "embed"
    },
    "jina-clip-v1": {
        "source": {
            "huggingface": "jinaai/jina-clip-v1",
            "modelscope": "jinaai/jina-clip-v1"
        },
        "type": "cross_modal_embed" # image - text
    },
    "siglip": {
        "source": {
            "huggingface": "google/siglip-so400m-patch14-384",
        },
        "type": "cross_modal_embed" # image - text
    },
    "colqwen2-v0.1": {
        "source": {
            "huggingface": "michaelfeil/colqwen2-v0.1",
        },
        "type": "cross_modal_embed" # image - text
    },
    "bge-reranker-large": {
        "source": {
            "huggingface": "BAAI/bge-reranker-large",
            "modelscope": "Xorbits/bge-reranker-large"
        },
        "type": "reranker"
    },
    "chattts": {
        "source": {
            "huggingface": "2Noise/ChatTTS",
            "modelscope": "AI-ModelScope/ChatTTS"
        },
        "type": "tts"
    },
    "internvl-chat-v1-5": {
        "source": {
            "huggingface": "OpenGVLab/InternVL-Chat-V1-5",
            "modelscope": "OpenGVLab/InternVL-Chat-V1-5"
        },
        "type": "vlm"
    },
    "yi-vl-6b": {
        "source": {
            "huggingface": "01-ai/Yi-VL-6B",
            "modelscope": "01ai/Yi-VL-6B"
        },
        "type": "vlm"
    },
    "llava-1.5-13b-hf": {
        "source": {
            "huggingface": "llava-hf/llava-1.5-13b-hf",
            "modelscope": "huangjintao/llava-1.5-13b-hf"
        },
        "type": "vlm"
    },
    "llava-1.5-7b-hf": {
        "source": {
            "huggingface": "llava-hf/llava-1.5-7b-hf",
            "modelscope": "huangjintao/llava-1.5-7b-hf"
        },
        "type": "vlm"
    },
    "mini-internvl-chat-2b-v1-5": {
        "source": {
            "huggingface": "OpenGVLab/Mini-InternVL-Chat-2B-V1-5",
            "modelscope": "OpenGVLab/Mini-InternVL-Chat-2B-V1-5"
        },
        "type": "vlm"
    },
    "mini-internvl-chat-4b-v1-5": {
        "source": {
            "huggingface": "OpenGVLab/Mini-InternVL-Chat-4B-V1-5",
            "modelscope": "OpenGVLab/Mini-InternVL-Chat-4B-V1-5"
        },
        "type": "vlm"
    },
    "musicgen-medium": {
        "source": {
            "huggingface": "facebook/musicgen-medium",
            "modelscope": "AI-ModelScope/musicgen-medium"
        },
        "type": "tts"
    },
    "musicgen-stereo-small": {
        "source": {
            "huggingface": "facebook/musicgen-stereo-small",
            "modelscope": "AI-ModelScope/musicgen-small"
        },
        "type": "tts"
    },
    "sensevoicesmall": {
        "source": {
            "huggingface": "FunAudioLLM/SenseVoiceSmall",
            "modelscope": "iic/SenseVoiceSmall"
        },
        "type": "stt"
    },
    "stable-diffusion-3-medium": {
        "source": {
            "huggingface": "stabilityai/stable-diffusion-3-medium",
            "modelscope": "AI-ModelScope/stable-diffusion-3-medium-diffusers"
        },
        "type": "sd"
    }
}



import os
import json
from datetime import datetime
from dataclasses import dataclass, asdict

import lazyllm
from lazyllm.thirdparty import datasets
from ..components.utils.file_operate import delete_old_files

@dataclass
class TrainConfig:
    finetune_model_name: str = 'llm'
    base_model: str = 'llm'
    training_type: str = 'SFT'
    finetuning_type: str = 'LoRA'
    data_path: str = 'path/to/dataset'
    num_gpus: int = 1
    val_size: float = 0.1
    num_epochs: int = 1
    learning_rate: float = 1e-4
    lr_scheduler_type: str = 'cosine'
    batch_size: int = 32
    cutoff_len: int = 1024
    lora_r: int = 8
    lora_alpha: int = 32
    lora_rate: float = 0.1

def update_config(input_dict: dict, default_data: type) -> dict:
    config = TrainConfig()
    config_dict = asdict(config)
    assert all([key in config_dict for key in input_dict.keys()]), \
        f"The {input_dict.keys()} must be the subset of {config_dict.keys()}."
    config_dict.update(input_dict)
    return config_dict

INPUT_SPLIT = " ### input "

def uniform_sft_dataset(dataset_path: str, target: str = 'alpaca') -> str:
    '''
    {origin_format}.{suffix} -> {target_format}, supported all 8 cases:
    1. openai.json   -> alpaca: Conversion: openai2alpaca: json
    2. openai.jsonl  -> alpaca: Conversion: openai2alpaca: json
    3. alpaca.json   -> alpaca: Keep: json
    4. alpaca.jsonl  -> alpaca: Restore: jsonl -> json
    5. openai.json   -> openai: Restore: json -> jsonl
    6. openai.jsonl  -> openai: Keep: jsonl
    7. alpaca.json   -> openai: Conversion: alpaca2openai: jsonl
    8. alpaca.jsonl  -> openai: Conversion: alpaca2openai: jsonl
    Note: target-suffix does match:{'openai': 'jsonl'; 'alpaca': 'json'}
    '''
    assert os.path.exists(dataset_path), f"Path: {dataset_path} does not exist!"

    data = datasets.load_dataset('json', data_files=dataset_path)
    file_name = os.path.basename(dataset_path)
    base_name, suffix = file_name.split('.')
    assert suffix in ['json', 'jsonl']
    target = target.strip().lower()
    save_suffix = 'json'

    # Get the format('alpaca' or 'openai') of the original dataset
    origin_format = 'alpaca'
    if "messages" in data["train"][0]:
        origin_format = 'openai'

    # Verify that the dataset format is consistent with the target format
    if origin_format == target:
        if target == 'alpaca':
            if suffix == 'json':
                return dataset_path
            else:
                save_data = alpaca_filter_null(data)
        else:
            if suffix == 'jsonl':
                return dataset_path
            else:
                save_suffix = 'jsonl'
                save_data = data['train'].to_list()
    else:
        # The format is inconsistent, conversion is required
        if target == 'alpaca':
            save_data = openai2alpaca(data)
        elif target == 'openai':
            save_data = alpaca2openai(data)
            save_suffix = 'jsonl'
        else:
            raise ValueError(f"Not supported type: {target}")

    return save_dataset(save_data, save_suffix, base_name + f'_{suffix}')

def save_json(data: list, output_json_path: str) -> None:
    with open(output_json_path, 'w', encoding='utf-8') as json_file:
        json.dump(data, json_file, ensure_ascii=False, indent=4)

def save_jsonl(data: list, output_json_path: str) -> None:
    with open(output_json_path, mode='w', encoding='utf-8') as json_file:
        for row in data:
            json_file.write(json.dumps(row, ensure_ascii=False) + '\n')

def save_dataset(save_data: list, save_suffix='json', base_name='train_data') -> str:
    directory = os.path.join(lazyllm.config['temp_dir'], 'dataset')
    if not os.path.exists(directory):
        os.makedirs(directory)
    delete_old_files(directory)
    time_stamp = datetime.now().strftime('%y%m%d%H%M%S%f')[:14]
    output_json_path = os.path.join(directory, f'{base_name}_{time_stamp}.{save_suffix}')
    if save_suffix == 'json':
        save_json(save_data, output_json_path)
    else:
        save_jsonl(save_data, output_json_path)
    return output_json_path

def alpaca_filter_null(data) -> list:
    res = []
    for item in data["train"]:
        alpaca_item = dict()
        for key in item.keys():
            if item[key]:
                alpaca_item[key] = item[key]
        res.append(alpaca_item)
    return res

def alpaca2openai(data) -> list:
    res = []
    for item in data["train"]:
        openai_item = {"messages": []}
        inp = item.get("input", "")
        system = item.get("system", "")  # Maybe get None
        historys = item.get("history", [])
        if system:
            openai_item["messages"].append({"role": "system", "content": system})
        openai_item["messages"].extend([
            {"role": "user", "content": item["instruction"] + (INPUT_SPLIT + inp if inp else "")},
            {"role": "assistant", "content": item["output"]}
        ])
        if historys:
            for history in historys:
                openai_item["messages"].append({"role": "user", "content": history[0]})
                openai_item["messages"].append({"role": "assistant", "content": history[1]})

        res.append(openai_item)

    return res

def openai2alpaca(data) -> list:
    res = []
    for line in data["train"]:
        chat = line["messages"]
        system = ''
        instructions = []
        outputs = []
        for item in chat:
            if item["role"] == "system" and not system:
                system = item["content"]
            if item["role"] == "user":
                instructions.append(item["content"])
            if item["role"] == "assistant":
                outputs.append(item["content"])
        assert len(instructions) == len(outputs)
        history = [[x, y] for x, y in zip(instructions[1:], outputs[1:])]
        instruction_input = instructions[0].split(INPUT_SPLIT)
        instruction = instruction_input[0]
        inp = ''
        if len(instruction_input) >= 2:
            inp = instruction_input[-1]
        output = outputs[0]
        alpaca_item = dict()
        if system:
            alpaca_item["system"] = system
        alpaca_item["instruction"] = instruction
        # fixed llama-factory-bug: must have input
        alpaca_item["input"] = inp
        alpaca_item["output"] = output
        if history:
            alpaca_item["history"] = history
        res.append(alpaca_item)
    return res



from lazyllm.module import ModuleBase
from lazyllm.components import AlpacaPrompter
from lazyllm import pipeline, globals, switch
from lazyllm.tools.utils import chat_history_to_str
from typing import Dict, Union, Any, List
import json


en_prompt_classifier_template = """
## role：Intent Classifier
You are an intent classification engine responsible for analyzing user input text based on dialogue information and determining a unique intent category.{user_prompt}

## Constrains:
You only need to reply with the name of the intent. Do not output any additional fields and do not translate it.{user_constrains}

## Attention:
{attention}

## Text Format
The input text is in JSON format, where "human_input" contains the user's raw input and "intent_list" contains a list of all intent names.

## Example
User: {{"human_input": "What’s the weather like in Beijing tomorrow?", "intent_list": ["Check Weather", "Search Engine Query", "View Surveillance", "Report Summary", "Chat"]}}
Assistant: Check Weather
{user_examples}

## Conversation History
The chat history between the human and the assistant is stored within the <histories></histories> XML tags below.
<histories>
{history_info}
</histories>

Input text is as follows:
"""  # noqa E501

ch_prompt_classifier_template = """
## role：意图分类器
你是一个意图分类引擎，负责根据对话信息分析用户输入文本并确定唯一的意图类别。{user_prompt}

## 限制:
你只需要回复意图的名字即可，不要额外输出其他字段，也不要进行翻译。{user_constrains}

## 注意:
{attention}

## 文本格式
输入文本为JSON格式，"human_input"中内容为用户的原始输入，"intent_list"为所有意图名列表

## 示例
User: {{"human_input": "北京明天天气怎么样？", "intent_list": ["查看天气", "搜索引擎检索", "查看监控", "周报总结", "聊天"]}}
Assistant:  查看天气
{user_examples}

## 历史对话
人类和助手之间的聊天记录存储在下面的 <histories></histories> XML 标记中。
<histories>
{history_info}
</histories>

输入文本如下:
"""  # noqa E501


class IntentClassifier(ModuleBase):
    """IntentClassifier is an intent recognizer based on a language model that identifies predefined intents based on user-provided input text and conversational context.
It can handle intent lists and ensures accurate intent recognition through preprocessing and postprocessing steps.

Arguments:
    llm: A language model object used for intent recognition, which can be of type OnlineChatModule or TrainableModule.
    intent_list (list): A list of strings containing all possible intents. This list can include intents in either Chinese or English.
    prompt (str): User-attached prompt words.
    constrain (str): User-attached constrain words.
    examples (list[list]): extra examples，format is `[[query, intent], [query, intent], ...]`.
    return_trace (bool, optional): If set to True, the results will be recorded in the trace. Defaults to False.


Examples:
        >>> import lazyllm
        >>> from lazyllm.tools import IntentClassifier
        >>> classifier_llm = lazyllm.OnlineChatModule(source="openai")
        >>> chatflow_intent_list = ["Chat", "Financial Knowledge Q&A", "Employee Information Query", "Weather Query"]
        >>> classifier = IntentClassifier(classifier_llm, intent_list=chatflow_intent_list)
        >>> classifier.start()
        >>> print(classifier('What is the weather today'))
        Weather Query
        >>>
        >>> with IntentClassifier(classifier_llm) as ic:
        >>>     ic.case['Weather Query', lambda x: '38.5°C']
        >>>     ic.case['Chat', lambda x: 'permission denied']
        >>>     ic.case['Financial Knowledge Q&A', lambda x: 'Calling Financial RAG']
        >>>     ic.case['Employee Information Query', lambda x: 'Beijing']
        ...
        >>> ic.start()
        >>> print(ic('What is the weather today'))
        38.5°C
    """
    def __init__(self, llm, intent_list: list = None,
                 *, prompt: str = '', constrain: str = '', attention: str = '',
                 examples: list[list[str, str]] = [], return_trace: bool = False) -> None:
        super().__init__(return_trace=return_trace)
        self._intent_list = intent_list or []
        self._llm = llm
        self._prompt, self._constrain, self._attention, self._examples = prompt, constrain, attention, examples
        if self._intent_list:
            self._init()

    def _init(self):
        def choose_prompt():
            # Use chinese prompt if intent elements have chinese character, otherwise use english version
            for ele in self._intent_list:
                for ch in ele:
                    # chinese unicode range
                    if "\u4e00" <= ch <= "\u9fff":
                        return ch_prompt_classifier_template
            return en_prompt_classifier_template

        example_template = '\nUser: {{{{"human_input": "{inp}", "intent_list": {intent}}}}}\nAssistant: {label}\n'
        examples = ''.join([example_template.format(
            inp=input, intent=self._intent_list, label=label) for input, label in self._examples])
        prompt = choose_prompt().replace(
            '{user_prompt}', f' {self._prompt}').replace('{attention}', self._attention).replace(
            '{user_constrains}', f' {self._constrain}').replace('{user_examples}', f' {examples}')
        self._llm = self._llm.share(prompt=AlpacaPrompter(dict(system=prompt, user='${input}')
                                                          ).pre_hook(self.intent_promt_hook)).used_by(self._module_id)
        self._impl = pipeline(self._llm, self.post_process_result)

    def intent_promt_hook(
        self,
        input: Union[str, List, Dict[str, str], None] = None,
        history: List[Union[List[str], Dict[str, Any]]] = [],
        tools: Union[List[Dict[str, Any]], None] = None,
        label: Union[str, None] = None,
    ):
        input_json = {}
        if isinstance(input, str):
            input_json = {"human_input": input, "intent_list": self._intent_list}
        else:
            raise ValueError(f"Unexpected type for input: {type(input)}")

        history_info = chat_history_to_str(history)
        history = []
        input_text = json.dumps(input_json, ensure_ascii=False)
        return dict(history_info=history_info, input=input_text), history, tools, label

    def post_process_result(self, input):
        input = input.strip()
        return input if input in self._intent_list else self._intent_list[0]

    def forward(self, input: str, llm_chat_history: List[Dict[str, Any]] = None):
        if llm_chat_history is not None and self._llm._module_id not in globals["chat_history"]:
            globals["chat_history"][self._llm._module_id] = llm_chat_history
        return self._impl(input)

    def __enter__(self):
        assert not self._intent_list, 'Intent list is already set'
        self._sw = switch()
        self._sw.__enter__()
        return self

    @property
    def case(self):
        return switch.Case(self)

    @property
    def submodules(self):
        submodule = []
        if isinstance(self._impl, switch):
            self._impl.for_each(lambda x: isinstance(x, ModuleBase), lambda x: submodule.append(x))
        return super().submodules + submodule

    # used by switch.Case
    def _add_case(self, cond, func):
        assert isinstance(cond, str), 'intent must be string'
        self._intent_list.append(cond)
        self._sw.case[cond, func]

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._sw.__exit__(exc_type, exc_val, exc_tb)
        self._init()
        self._sw._set_conversion(self._impl)
        self._impl = self._sw



import re
from pathlib import Path
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem
from typing import Dict, List, Optional, Tuple

from .readerBase import LazyLLMReaderBase
from ..doc_node import DocNode

class MarkdownReader(LazyLLMReaderBase):
    def __init__(self, remove_hyperlinks: bool = True, remove_images: bool = True, return_trace: bool = True) -> None:
        super().__init__(return_trace=return_trace)
        self._remove_hyperlinks = remove_hyperlinks
        self._remove_images = remove_images

    def _markdown_to_tups(self, markdown_text: str) -> List[Tuple[Optional[str], str]]:
        markdown_tups: List[Tuple[Optional[str], str]] = []
        lines = markdown_text.split("\n")

        current_header = None
        current_lines = []
        in_code_block = False

        for line in lines:
            if line.startswith("```"): in_code_block = not in_code_block

            header_match = re.match(r"^#+\s", line)
            if not in_code_block and header_match:
                if current_header is not None or len(current_lines) > 0:
                    markdown_tups.append((current_header, "\n".join(current_lines)))
                current_header = line
                current_lines.clear()
            else:
                current_lines.append(line)

        markdown_tups.append((current_header, "\n".join(current_lines)))
        return [(key if key is None else re.sub(r"#", "", key).strip(), re.sub(r"<.*?>", "", value),)
                for key, value in markdown_tups]

    def remove_images(self, content: str) -> str:
        pattern = r"!{1}\[\[(.*)\]\]"
        return re.sub(pattern, "", content)

    def remove_hyperlinks(self, content: str) -> str:
        pattern = r"\[(.*)\]\((.*)\)"
        return re.sub(pattern, r"\1", content)

    def _parse_tups(self, filepath: Path, errors: str = "ignore",
                    fs: Optional[AbstractFileSystem] = None) -> List[Tuple[Optional[str], str]]:
        fs = fs or LocalFileSystem()

        with fs.open(filepath, encoding="utf-8") as f:
            content = f.read().decode(encoding="utf-8")

        if self._remove_hyperlinks: content = self.remove_hyperlinks(content)
        if self._remove_images: content = self.remove_images(content)
        return self._markdown_to_tups(content)

    def _load_data(self, file: Path, extra_info: Optional[Dict] = None,
                   fs: Optional[AbstractFileSystem] = None) -> List[DocNode]:
        if not isinstance(file, Path): file = Path(file)

        tups = self._parse_tups(file, fs=fs)
        results = [DocNode(
            content=[value if header is None else f"\n\n{header}\n{value}" for header, value in tups],
            global_metadata=extra_info)]
        return results



import os
import uuid
from lazyllm.thirdparty import scipy, numpy as np
from ..utils.file_operate import delete_old_files
import lazyllm
from lazyllm import LOG


def sound_to_file(sound: 'np.array', file_path: str, sample_rate: int = 24000) -> str:
    scaled_audio = np.int16(sound / np.max(np.abs(sound)) * 32767)
    scipy.io.wavfile.write(file_path, sample_rate, scaled_audio)
    return [file_path]

def sounds_to_files(sounds: list, directory: str, sample_rate: int = 24000) -> list:
    if not os.path.exists(directory):
        os.makedirs(directory)
    delete_old_files(directory)
    unique_id = uuid.uuid4()
    path_list = []
    for i, sound in enumerate(sounds):
        file_path = os.path.join(directory, f'sound_{unique_id}_{i}.wav')
        sound_to_file(sound, file_path, sample_rate)
        path_list.append(file_path)
    return path_list


class TTSBase(object):
    func = None

    def __init__(self, launcher=None, log_path=None):
        self.launcher = launcher
        self._log_path = log_path

    def __call__(self, finetuned_model=None, base_model=None):
        if not finetuned_model:
            finetuned_model = base_model
        elif not os.path.exists(finetuned_model) or \
            not any(file.endswith(('.bin', '.safetensors'))
                    for _, _, filenames in os.walk(finetuned_model) for file in filenames):
            LOG.warning(f"Note! That finetuned_model({finetuned_model}) is an invalid path, "
                        f"base_model({base_model}) will be used")
            finetuned_model = base_model
        return lazyllm.deploy.RelayServer(func=self.__class__.func(finetuned_model), launcher=self.launcher,
                                          log_path=self._log_path, cls='tts')()



import re
import json
import collections
from lazyllm import LOG

templates = dict(
    # Template used by Alpaca-LoRA.
    alpaca={
        "prompt": "Below is an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.\n\n### Instruction:\n{instruction}\n\n### Input:\n{input}\n\n### Response:\n",  # noqa E501
        "response_split": "### Response:",
    },
    # Template used by internLM
    puyu={
        "prompt": "<bos><|System|>:You are an AI assistant whose name is InternLM (书生·浦语).\n- InternLM (书生·浦语) is a conversational language model that is developed by Shanghai AI Laboratory (上海人工智能实验室). It is designed to be helpful, honest, and harmless.\n- InternLM (书生·浦语) can understand and communicate fluently in the language chosen by the user such as English and 中文.<eosys>\n<|Human|>:{instruction}<eoh>\n<|Assistant|>:please tell me what to do。ി\n<|Human|>:{input}<eoh>\n<|Assistant|>:", # noqa E501
        "response_split": "<|Assistant|>:",
    }
)

class Prompter(object):
    def __init__(self, prompt=None, response_split=None, *, chat_prompt=None,
                 history_symbol='llm_chat_history', eoa=None, eoh=None, show=False):
        self._prompt, self._response_split = prompt, response_split
        self._chat_prompt = chat_prompt
        self._history_symbol, self._eoa, self._eoh = history_symbol, eoa, eoh
        self._show = show
        self._prompt_keys = list(set(re.findall(r'\{(\w+)\}', self._prompt))) if prompt else []
        if chat_prompt is not None:
            chat_keys = set(re.findall(r'\{(\w+)\}', self._chat_prompt))
            assert set(self._prompt_keys).issubset(chat_keys)
            assert chat_keys - set(self._prompt_keys) == set([self._history_symbol])
            self.use_history = True
        else:
            self.use_history = history_symbol in self._prompt_keys
            if self.use_history:
                self._prompt_keys.pop(self._prompt_keys.index(history_symbol))
                self._chat_prompt = self._prompt

    @classmethod
    def from_dict(cls, prompt, *, show=False):
        assert isinstance(prompt, dict)
        return cls(**prompt, show=show)

    @classmethod
    def from_template(cls, template_name, *, show=False):
        return cls.from_dict(templates[template_name], show=show)

    @classmethod
    def from_file(cls, fname, *, show=False):
        with open(fname) as fp:
            return cls.from_dict(json.load(fp), show=show)

    @classmethod
    def empty(cls):
        return cls()

    def is_empty(self):
        return self._prompt is None

    def generate_prompt(self, input, history=None, tools=None, label=None, show=False):
        if not self.is_empty():
            assert tools is None
            # datasets.formatting.formatting.LazyDict is used in transformers
            if not isinstance(input, collections.abc.Mapping):
                assert len(self._prompt_keys) == 1, (
                    f'invalid prompt `{self._prompt}` for <{type(input)}> input `{input}`')
                input = {self._prompt_keys[0]: input}
            try:
                if self.use_history and isinstance(history, list) and len(history) > 0:
                    assert isinstance(history[0], list), 'history must be list of list'
                    input[self._history_symbol] = self._eoa.join([self._eoh.join(h) for h in history])
                    input = self._chat_prompt.format(**input)
                else:
                    if self.use_history: input[self._history_symbol] = ''
                    input = self._prompt.format(**input)
            except Exception:
                raise RuntimeError(f'Generate prompt failed, and prompt is {self._prompt}; chat-prompt'
                                   f' is {self._chat_prompt}; input is {input}; history is {history}')
            if label: input += label
        if self._show or show: LOG.info(input)
        return input

    def get_response(self, response, input=None):
        if input and response.startswith(input):
            return response[len(input):]
        return response if self._response_split is None else response.split(self._response_split)[-1]



import lazyllm
from urllib.parse import urljoin
from .onlineChatModuleBase import OnlineChatModuleBase

class KimiModule(OnlineChatModuleBase):

    def __init__(self,
                 base_url: str = "https://api.moonshot.cn/",
                 model: str = "moonshot-v1-8k",
                 api_key: str = None,
                 stream: bool = True,
                 return_trace: bool = False,
                 **kwargs):

        super().__init__(model_series="KIMI",
                         api_key=api_key or lazyllm.config['kimi_api_key'],
                         base_url=base_url,
                         model_name=model,
                         stream=stream,
                         trainable_models=[],
                         return_trace=return_trace,
                         **kwargs)

    def _get_system_prompt(self):
        return ("You are Kimi, an AI assistant provided by Moonshot AI. You are better at speaking "
                "Chinese and English. You will provide users with safe, helpful, and accurate answers. "
                "At the same time, you will reject all answers involving terrorism, racial discrimination, "
                "pornographic violence, etc. Moonshot AI is a proper noun and cannot be translated "
                "into other languages.")

    def _set_chat_url(self):
        self._url = urljoin(self._base_url, 'v1/chat/completions')

    def _format_vl_chat_image_url(self, image_url, mime):
        assert not image_url.startswith("http"), "Kimi vision model only supports base64 format"
        assert mime is not None, "Kimi Module requires mime info."
        image_url = f"data:{mime};base64,{image_url}"
        return [{"type": "image_url", "image_url": {"url": image_url}}]

    def _format_vl_chat_query(self, query: str):
        return query



# flake8: noqa
from collections.abc import Callable, Generator, Iterator
import csv
import itertools
from core import protocol as proto
import typing
import unittest
from typing import List, Dict, Union

T = typing.TypeVar("T")

def values(rule: proto.Rule[T], default: Union[T, None] = None) -> List[T]:
    if rule.options is None:
        return [default or rule.value_type()]
    return rule.options

def overwrite(input: Dict[str, List[typing.Any]], **kwargs: List[typing.Any]) -> Dict[str, List[typing.Any]]:
    for key, value in kwargs.items():
        assert key in input
        assert set(value) <= set(input[key])
    input.update(**kwargs)
    return input

def combine(options: Dict[str, List[typing.Any]], **constraints: Callable[[Dict[str, typing.Any]], bool])\
        -> Generator[Iterator[typing.Any], typing.Any, None]:
    limit = len(options)
    input = [(key, value) for key, value in options.items()]
    output: Dict[str, typing.Any] = {}

    def f(index: int) -> Generator[Iterator[typing.Any], typing.Any, None]:
        if index == limit:
            yield output.values().__iter__()
        else:
            name, items = input[index]
            for item in items:
                output[name] = item
                if constraints.get(name, lambda _: True)(output):
                    yield from f(index + 1)
                output[name] = None

    yield from f(0)


def generate(filename: str, head: List[typing.Any], body: Iterator[Iterator[typing.Any]]):
    with open(filename, 'w', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(head)
        count = 0
        for value in body:
            writer.writerow(value)
            count += 1
        return count


class TestGenerateEmptyCSV(unittest.TestCase):
    def test_generate_csv(self): # noqa C901
        rules = proto.TRAINING_RULE_SET
        items = {rule.name: values(rule) for rule in rules}
        items = overwrite(
            items,
            GPU_NUM=[1, 2, 4, 8, 16, 24, 32],
            MODEL_NAME=["LLAMA2_7B", "LLAMA2_13B", "LLAMA2_70B"],
            CTX_LEN=[128, 256, 512, 1024, 2048, 4096],
            BATCH_SIZE=[16],
            TRAINABLE_PARAMS=[0],
            ZERO=[True],
            LORA_R=[0, 8, 16, 32])

        def to_parameter_size(model_name: str) -> int:
            return {
                "LLAMA_7B": 7,
                "LLAMA_13B": 13,
                "LLAMA_20B": 20,
                "LLAMA_65B": 65,
                "LLAMA2_7B": 7,
                "LLAMA2_13B": 13,
                "LLAMA2_70B": 70,
            }[model_name]

        def to_gpu_memory(x: str) -> int:
            return {"A100": 80, "A800": 80}[x]

        def is_inference_framework(x: str) -> bool:
            return x in ["LIGHTLLM", "VLLM"]

        def model_name_constraint(_: Dict[str, typing.Any]) -> bool:
            mapping = {
                "LLAMA_7B": 80,
                "LLAMA_13B": 80 * 2,
                "LLAMA_20B": 80 * 2,
                "LLAMA_65B": 80 * 4,
                "LLAMA2_7B": 80,
                "LLAMA2_13B": 80 * 2,
                "LLAMA2_70B": 80 * 4,
            }
            r = mapping[_["MODEL_NAME"]]
            m = to_gpu_memory(_["GPU_TYPE"])
            options = proto.GPU_NUM.options
            assert options is not None and proto.GPU_NUM.indexed
            le = min(i for i, x in enumerate(options) if r <= x * m)
            r = options.index(_["GPU_NUM"])
            return le <= r <= le + 2

        def tp_constraint(_: Dict[str, typing.Any]) -> bool:
            if _["TP"] > _["GPU_NUM"] or to_parameter_size(_["MODEL_NAME"]) > 30 * _["TP"]:
                return False
            if _["FRAMEWORK"] in ["ALPACA"]:
                return _["TP"] == min(_["GPU_NUM"], 8)
            return True

        def pp_constraint(_: Dict[str, typing.Any]) -> bool:
            if _["TP"] * _["PP"] > _["GPU_NUM"]:
                return False
            if _["FRAMEWORK"] in ["ALPACA"]:
                return _["PP"] == 1
            return True

        def gradient_step_constraint(_: Dict[str, typing.Any]) -> bool:
            options = proto.GRADIENT_STEP.options
            gradient_size = {
                "LLAMA_7B": 4,
                "LLAMA_13B": 4,
                "LLAMA_20B": 4,
                "LLAMA_65B": 4,
                "LLAMA2_7B": 4,
                "LLAMA2_13B": 4,
                "LLAMA2_70B": 4,
            }
            minimium_activation_size = {
                "LLAMA_7B": 3,
                "LLAMA_13B": 5,
                "LLAMA_20B": 5,
                "LLAMA_65B": 6,
                "LLAMA2_7B": 3,
                "LLAMA2_13B": 5,
                "LLAMA2_70B": 6,
            }

            assert options is not None
            for i, gradient_step in enumerate(options):
                require = to_parameter_size(_["MODEL_NAME"]) * 2 / _["TP"] + \
                    gradient_size[_["MODEL_NAME"]] * _["GRADIENT_STEP"] + \
                    minimium_activation_size[_["MODEL_NAME"]] * _["CTX_LEN"] * \
                    _["BATCH_SIZE"] * _["TP"] * _["PP"] / gradient_step / 128.0
                if require < to_gpu_memory(_["GPU_TYPE"]) * _["GPU_NUM"]:
                    return _["GRADIENT_STEP"] in options[i: min(i + 1, len(options))]
            return False

        def lora_r_constraint(_: Dict[str, typing.Any]) -> bool:
            if is_inference_framework(_["FRAMEWORK"]):
                return _["LORA_R"] == 0
            return _["LORA_R"] > 0

        body = combine(items,
                       MODEL_NAME=model_name_constraint,
                       TP=tp_constraint,
                       PP=pp_constraint,
                       GRADIENT_STEP=gradient_step_constraint,
                       LORA_R=lora_r_constraint)

        head = [rule.name for rule in rules]
        size = generate('sample_finetune_empty.csv', head, itertools.islice(body, 1000000))
        print(f"GENERATE {size} cases")


if __name__ == '__main__':
    unittest.main()



from typing import Any, Dict, List
import lazyllm
from .onlineEmbeddingModuleBase import OnlineEmbeddingModuleBase

class QwenEmbedding(OnlineEmbeddingModuleBase):

    def __init__(self,
                 embed_url: str = ("https://dashscope.aliyuncs.com/api/v1/services/"
                                   "embeddings/text-embedding/text-embedding"),
                 embed_model_name: str = "text-embedding-v1",
                 api_key: str = None):
        super().__init__("QWEN", embed_url, api_key or lazyllm.config['qwen_api_key'], embed_model_name)

    def _encapsulated_data(self, text: str, **kwargs) -> Dict[str, str]:
        json_data = {
            "input": {
                "texts": [text]
            },
            "model": self._embed_model_name
        }
        if len(kwargs) > 0:
            json_data.update(kwargs)

        return json_data

    def _parse_response(self, response: Dict[str, Any]) -> List[float]:
        return response['output']['embeddings'][0]['embedding']


class QwenReranking(OnlineEmbeddingModuleBase):

    def __init__(self,
                 embed_url: str = ("https://dashscope.aliyuncs.com/api/v1/services/"
                                   "rerank/text-rerank/text-rerank"),
                 embed_model_name: str = "gte-rerank",
                 api_key: str = None, **kwargs):
        super().__init__("QWEN", embed_url, api_key or lazyllm.config['qwen_api_key'], embed_model_name)

    @property
    def type(self):
        return "ONLINE_RERANK"

    def _encapsulated_data(self, query: str, documents: List[str], top_n: int, **kwargs) -> Dict[str, str]:
        json_data = {
            "input": {
                "query": query,
                "documents": documents
            },
            "parameters": {
                "top_n": top_n,
            },
            "model": self._embed_model_name
        }
        if len(kwargs) > 0:
            json_data.update(kwargs)

        return json_data

    def _parse_response(self, response: Dict[str, Any]) -> List[float]:
        results = response['output']['results']
        return [(result["index"], result["relevance_score"]) for result in results]



import json
from .formatterbase import JsonLikeFormatter
import lazyllm

class JsonFormatter(JsonLikeFormatter):
    """This class is a JSON formatter, that is, the user wants the model to output content is JSON format, and can also select a field in the output content by indexing.


Examples:
    >>> import lazyllm
    >>> from lazyllm.components import JsonFormatter
    >>> toc_prompt='''
    ... You are now an intelligent assistant. Your task is to understand the user's input and convert the outline into a list of nested dictionaries. Each dictionary contains a `title` and a `describe`, where the `title` should clearly indicate the level using Markdown format, and the `describe` is a description and writing guide for that section.
    ... 
    ... Please generate the corresponding list of nested dictionaries based on the following user input:
    ... 
    ... Example output:
    ... [
    ...     {
    ...         "title": "# Level 1 Title",
    ...         "describe": "Please provide a detailed description of the content under this title, offering background information and core viewpoints."
    ...     },
    ...     {
    ...         "title": "## Level 2 Title",
    ...         "describe": "Please provide a detailed description of the content under this title, giving specific details and examples to support the viewpoints of the Level 1 title."
    ...     },
    ...     {
    ...         "title": "### Level 3 Title",
    ...         "describe": "Please provide a detailed description of the content under this title, deeply analyzing and providing more details and data support."
    ...     }
    ... ]
    ... User input is as follows:
    ... '''
    >>> query = "Please help me write an article about the application of artificial intelligence in the medical field."
    >>> m = lazyllm.TrainableModule("internlm2-chat-20b").prompt(toc_prompt).start()
    >>> ret = m(query, max_new_tokens=2048)
    >>> print(f"ret: {ret!r}")  # the model output without specifying a formatter
    'Based on your user input, here is the corresponding list of nested dictionaries:
    [
        {
            "title": "# Application of Artificial Intelligence in the Medical Field",
            "describe": "Please provide a detailed description of the application of artificial intelligence in the medical field, including its benefits, challenges, and future prospects."
        },
        {
            "title": "## AI in Medical Diagnosis",
            "describe": "Please provide a detailed description of how artificial intelligence is used in medical diagnosis, including specific examples of AI-based diagnostic tools and their impact on patient outcomes."
        },
        {
            "title": "### AI in Medical Imaging",
            "describe": "Please provide a detailed description of how artificial intelligence is used in medical imaging, including the advantages of AI-based image analysis and its applications in various medical specialties."
        },
        {
            "title": "### AI in Drug Discovery and Development",
            "describe": "Please provide a detailed description of how artificial intelligence is used in drug discovery and development, including the role of AI in identifying potential drug candidates and streamlining the drug development process."
        },
        {
            "title": "## AI in Medical Research",
            "describe": "Please provide a detailed description of how artificial intelligence is used in medical research, including its applications in genomics, epidemiology, and clinical trials."
        },
        {
            "title": "### AI in Genomics and Precision Medicine",
            "describe": "Please provide a detailed description of how artificial intelligence is used in genomics and precision medicine, including the role of AI in analyzing large-scale genomic data and tailoring treatments to individual patients."
        },
        {
            "title": "### AI in Epidemiology and Public Health",
            "describe": "Please provide a detailed description of how artificial intelligence is used in epidemiology and public health, including its applications in disease surveillance, outbreak prediction, and resource allocation."
        },
        {
            "title": "### AI in Clinical Trials",
            "describe": "Please provide a detailed description of how artificial intelligence is used in clinical trials, including its role in patient recruitment, trial design, and data analysis."
        },
        {
            "title": "## AI in Medical Practice",
            "describe": "Please provide a detailed description of how artificial intelligence is used in medical practice, including its applications in patient monitoring, personalized medicine, and telemedicine."
        },
        {
            "title": "### AI in Patient Monitoring",
            "describe": "Please provide a detailed description of how artificial intelligence is used in patient monitoring, including its role in real-time monitoring of vital signs and early detection of health issues."
        },
        {
            "title": "### AI in Personalized Medicine",
            "describe": "Please provide a detailed description of how artificial intelligence is used in personalized medicine, including its role in analyzing patient data to tailor treatments and predict outcomes."
        },
        {
            "title": "### AI in Telemedicine",
            "describe": "Please provide a detailed description of how artificial intelligence is used in telemedicine, including its applications in remote consultations, virtual diagnoses, and digital health records."
        },
        {
            "title": "## AI in Medical Ethics and Policy",
            "describe": "Please provide a detailed description of the ethical and policy considerations surrounding the use of artificial intelligence in the medical field, including issues related to data privacy, bias, and accountability."
        }
    ]'
    >>> m = lazyllm.TrainableModule("internlm2-chat-20b").formatter(JsonFormatter("[:][title]")).prompt(toc_prompt).start()
    >>> ret = m(query, max_new_tokens=2048)
    >>> print(f"ret: {ret}")  # the model output of the specified formaater
    ['# Application of Artificial Intelligence in the Medical Field', '## AI in Medical Diagnosis', '### AI in Medical Imaging', '### AI in Drug Discovery and Development', '## AI in Medical Research', '### AI in Genomics and Precision Medicine', '### AI in Epidemiology and Public Health', '### AI in Clinical Trials', '## AI in Medical Practice', '### AI in Patient Monitoring', '### AI in Personalized Medicine', '### AI in Telemedicine', '## AI in Medical Ethics and Policy']
    """
    def _extract_json_from_string(self, mixed_str: str):
        json_objects = []
        brace_level = 0
        current_json = ""
        in_string = False

        for char in mixed_str:
            if char == '"' and (len(current_json) == 0 or current_json[-1] != '\\'):
                in_string = not in_string

            if not in_string:
                if char == '{':
                    if brace_level == 0:
                        current_json = ""
                    brace_level += 1
                elif char == '}':
                    brace_level -= 1

            if brace_level > 0 or (brace_level == 0 and char == '}'):
                current_json += char

            if brace_level == 0 and current_json:
                try:
                    json.loads(current_json)
                    json_objects.append(current_json)
                    current_json = ""
                except json.JSONDecodeError:
                    continue

        return json_objects

    def _load(self, msg: str):
        # Convert str to json format
        assert msg.count("{") == msg.count("}"), f"{msg} is not a valid json string."
        try:
            json_strs = self._extract_json_from_string(msg)
            if len(json_strs) == 0:
                raise TypeError(f"{msg} is not a valid json string.")
            res = []
            for json_str in json_strs:
                res.append(json.loads(json_str))
            return res if len(res) > 1 else res[0]
        except Exception as e:
            lazyllm.LOG.info(f"Error: {e}")
            return ""



import json
import ast
from collections import defaultdict
from functools import wraps
from pathlib import Path
from typing import Callable, Dict, List, Optional, Set, Union, Tuple, Any
from lazyllm import LOG, once_wrapper
from .transform import (NodeTransform, FuncNodeTransform, SentenceSplitter, LLMParser,
                        AdaptiveTransform, make_transform, TransformArgs, TransformArgs as TArgs)
from .index_base import IndexBase
from .store_base import StoreBase, LAZY_ROOT_NAME, LAZY_IMAGE_GROUP
from .map_store import MapStore
from .chroma_store import ChromadbStore
from .milvus_store import MilvusStore
from .smart_embedding_index import SmartEmbeddingIndex
from .doc_node import DocNode
from .data_loaders import DirectoryReader
from .utils import DocListManager, gen_docid, is_sparse
from .global_metadata import GlobalMetadataDesc, RAG_DOC_ID, RAG_DOC_PATH
from .data_type import DataType
from dataclasses import dataclass
import threading
import time

_transmap = dict(function=FuncNodeTransform, sentencesplitter=SentenceSplitter, llm=LLMParser)

def embed_wrapper(func):
    if not func:
        return None

    @wraps(func)
    def wrapper(*args, **kwargs) -> List[float]:
        result = func(*args, **kwargs)
        return ast.literal_eval(result) if isinstance(result, str) else result

    return wrapper

class StorePlaceholder:
    pass

class EmbedPlaceholder:
    pass


class BuiltinGroups(object):
    @dataclass
    class Struct:
        name: str
        args: TransformArgs
        parent: str = LAZY_ROOT_NAME

        def __str__(self): return self.name

    CoarseChunk = Struct('CoarseChunk', TArgs(f=SentenceSplitter, kwargs=dict(chunk_size=1024, chunk_overlap=100)))
    MediumChunk = Struct('MediumChunk', TArgs(f=SentenceSplitter, kwargs=dict(chunk_size=256, chunk_overlap=25)))
    FineChunk = Struct('FineChunk', TArgs(f=SentenceSplitter, kwargs=dict(chunk_size=128, chunk_overlap=12)))


class DocImpl:
    _builtin_node_groups: Dict[str, Dict] = {}
    _global_node_groups: Dict[str, Dict] = {}
    _registered_file_reader: Dict[str, Callable] = {}

    def __init__(self, embed: Dict[str, Callable], dlm: Optional[DocListManager] = None,
                 doc_files: Optional[str] = None, kb_group_name: Optional[str] = None,
                 global_metadata_desc: Dict[str, GlobalMetadataDesc] = None,
                 store_conf: Optional[Dict] = None):
        super().__init__()
        assert (dlm is None) ^ (doc_files is None), 'Only one of dataset_path or doc_files should be provided'
        self._local_file_reader: Dict[str, Callable] = {}
        self._kb_group_name = kb_group_name or DocListManager.DEFAULT_GROUP_NAME
        self._dlm, self._doc_files = dlm, doc_files
        self._reader = DirectoryReader(None, self._local_file_reader, DocImpl._registered_file_reader)
        self.node_groups: Dict[str, Dict] = {LAZY_ROOT_NAME: {}, LAZY_IMAGE_GROUP: {}}
        self.embed = {k: embed_wrapper(e) for k, e in embed.items()}
        self._global_metadata_desc = global_metadata_desc
        self.store = store_conf  # NOTE: will be initialized in _lazy_init()
        self._activated_embeddings = {}
        self.index_pending_registrations = []

    @once_wrapper(reset_on_pickle=True)
    def _lazy_init(self) -> None:
        node_groups = DocImpl._builtin_node_groups.copy()
        node_groups.update(DocImpl._global_node_groups)
        node_groups.update(self.node_groups)
        self.node_groups = node_groups

        # set empty embed keys for groups that are not visited by Retriever
        for group in node_groups.keys():
            self._activated_embeddings.setdefault(group, set())

        embed_dims = {}
        embed_datatypes = {}
        for k, e in self.embed.items():
            embedding = e('a')
            if is_sparse(embedding):
                embed_datatypes[k] = DataType.SPARSE_FLOAT_VECTOR
            else:
                embed_dims[k] = len(embedding)
                embed_datatypes[k] = DataType.FLOAT_VECTOR

        if self.store is None:
            self.store = {
                'type': 'map',
            }

        if isinstance(self.store, Dict):
            self.store = self._create_store(store_conf=self.store, embed_dims=embed_dims,
                                            embed_datatypes=embed_datatypes)
        else:
            raise ValueError(f'store type [{type(self.store)}] is not a dict.')
        self._resolve_index_pending_registrations()

        if not self.store.is_group_active(LAZY_ROOT_NAME):
            ids, paths, metadatas = self._list_files()
            ids, paths, metadatas = self._delete_nonexistent_docs_on_startup(ids, paths, metadatas)
            if paths:
                if not metadatas: metadatas = [{} for _ in range(len(paths))]
                for idx, meta in enumerate(metadatas):
                    meta[RAG_DOC_ID] = ids[idx] if ids else gen_docid(paths[idx])
                    meta[RAG_DOC_PATH] = paths[idx]
                root_nodes = self._reader.load_data(paths, metadatas)
                self.store.update_nodes(root_nodes)
                if self._dlm:
                    self._dlm.update_kb_group(cond_file_ids=ids, cond_group=self._kb_group_name,
                                              new_status=DocListManager.Status.success)
                LOG.debug(f"building {LAZY_ROOT_NAME} nodes: {root_nodes}")
        if self._dlm:
            self._init_monitor_event = threading.Event()
            self._daemon = threading.Thread(target=self.worker)
            self._daemon.daemon = True
            self._daemon.start()
            self._init_monitor_event.wait()

    def _delete_nonexistent_docs_on_startup(self, ids, paths, metadatas):
        path_existing = [Path(path).exists() for path in paths]
        paths_need_delete = [paths[idx] for idx, exist in enumerate(path_existing) if not exist]
        rt_metadatas = [meta for meta, exist in zip(metadatas, path_existing) if exist] if metadatas else None
        rt_ids = [ids[idx] for idx, exist in enumerate(path_existing) if exist] if ids else None
        rt_paths = [path for path, exist in zip(paths, path_existing) if exist]

        if ids:
            ids_need_delete = [ids[idx] for idx, exist in enumerate(path_existing) if not exist]
        else:
            ids_need_delete = [gen_docid(path) for path in paths_need_delete]
        if ids_need_delete:
            if self._dlm is None:
                # if not using dlm, delete store directly;
                self._delete_doc_from_store(paths_need_delete)
            else:
                LOG.warning(f"Found {len(paths_need_delete)} docs that are not in store: {paths_need_delete}")
                # else dlm must turn on path monitoring to detect deleted files
                assert (self._dlm.enable_path_monitoring is True
                        ), 'DocListManager must turn on path monitoring or only use DocManager to delete files'
        return rt_ids, rt_paths, rt_metadatas

    def _resolve_index_pending_registrations(self):
        for index_type, index_cls, index_args, index_kwargs in self.index_pending_registrations:
            args = [self._resolve_index_placeholder(arg) for arg in index_args]
            kwargs = {k: self._resolve_index_placeholder(v) for k, v in index_kwargs.items()}
            self.store.register_index(index_type, index_cls(*args, **kwargs))
        self.index_pending_registrations.clear()

    def _create_store(self, store_conf: Optional[Dict], embed_dims: Optional[Dict[str, int]] = None,
                      embed_datatypes: Optional[Dict[str, DataType]] = None) -> StoreBase:
        store_type = store_conf.get('type')
        if not store_type:
            raise ValueError('store type is not specified.')

        kwargs = store_conf.get('kwargs', {})
        if not isinstance(kwargs, Dict):
            raise ValueError('`kwargs` in store conf is not a dict.')

        if store_type == "map":
            store = MapStore(node_groups=list(self._activated_embeddings.keys()), embed=self.embed, **kwargs)
        elif store_type == "chroma":
            store = ChromadbStore(group_embed_keys=self._activated_embeddings, embed=self.embed,
                                  embed_dims=embed_dims, **kwargs)
        elif store_type == "milvus":
            store = MilvusStore(group_embed_keys=self._activated_embeddings, embed=self.embed,
                                embed_dims=embed_dims, embed_datatypes=embed_datatypes,
                                global_metadata_desc=self._global_metadata_desc, **kwargs)
        else:
            raise NotImplementedError(
                f"Not implemented store type for {store_type}"
            )

        indices_conf = store_conf.get('indices', {})
        if not isinstance(indices_conf, Dict):
            raise ValueError(f"`indices`'s type [{type(indices_conf)}] is not a dict")

        for index_type, conf in indices_conf.items():
            if index_type == 'smart_embedding_index':
                backend_type = conf.get('backend')
                if not backend_type:
                    raise ValueError('`backend` is not specified in `smart_embedding_index`.')
                kwargs = conf.get('kwargs', {})
                index = SmartEmbeddingIndex(backend_type=backend_type,
                                            group_embed_keys=self._activated_embeddings,
                                            embed=self.embed,
                                            embed_dims=embed_dims,
                                            embed_datatypes=embed_datatypes,
                                            global_metadata_desc=self._global_metadata_desc,
                                            **kwargs)
            else:
                raise ValueError(f'unsupported index type [{index_type}]')

            store.register_index(type=index_type, index=index)

        return store

    @staticmethod
    def _create_node_group_impl(cls, group_name, name, transform: Union[str, Callable] = None,
                                parent: str = LAZY_ROOT_NAME, *, trans_node: bool = None,
                                num_workers: int = 0, **kwargs):
        group_name, parent = str(group_name), str(parent)
        groups = getattr(cls, group_name)

        def get_trans(t): return TransformArgs.from_dict(t) if isinstance(t, dict) else t

        if isinstance(transform, (TransformArgs, tuple, list, dict)):
            err_msg = '{} should be set in transform when transform is Dict, TransformArgs or List[TransformArgs]'
            assert trans_node is None, err_msg.format('trans_node')
            assert num_workers == 0, err_msg.format('num_workers')
            assert not kwargs, err_msg.format('kwargs')
            transforms = ([get_trans(t) for t in transform] if isinstance(transform, (list, tuple)) else
                          get_trans(transform))
        else:
            transforms = TransformArgs(f=transform, trans_node=trans_node,
                                       num_workers=num_workers, kwargs=kwargs)

        if name in groups:
            LOG.warning(f"Duplicate group name: {name}")
        for t in (transforms if isinstance(transform, list) else [transforms]):
            if isinstance(t.f, str):
                t.f = _transmap[t.f.lower()]
            if isinstance(t.f, type):
                assert t.trans_node is None, '`trans_node` is allowed only when transform is callable'
                if not issubclass(t.f, NodeTransform): LOG.warning(
                    'Please note! You are trying to use a completely custom transform class. The relationship '
                    'between nodes may become unreliable, `Document.get_parent/get_child` functions and the '
                    'target parameter of Retriever may have strange anomalies. Please use it at your own risk.')
            else:
                assert callable(t.f), f"transform should be callable, but get {t.f}"
        groups[name] = dict(transform=transforms, parent=parent)

    @classmethod
    def _create_builtin_node_group(cls, name, transform: Union[str, Callable] = None, parent: str = LAZY_ROOT_NAME,
                                   *, trans_node: bool = None, num_workers: int = 0, **kwargs) -> None:
        DocImpl._create_node_group_impl(cls, '_builtin_node_groups', name=name, transform=transform, parent=parent,
                                        trans_node=trans_node, num_workers=num_workers, **kwargs)

    @classmethod
    def create_global_node_group(cls, name, transform: Union[str, Callable] = None, parent: str = LAZY_ROOT_NAME,
                                 *, trans_node: bool = None, num_workers: int = 0, **kwargs) -> None:
        DocImpl._create_node_group_impl(cls, '_global_node_groups', name=name, transform=transform, parent=parent,
                                        trans_node=trans_node, num_workers=num_workers, **kwargs)

    def create_node_group(self, name, transform: Union[str, Callable] = None, parent: str = LAZY_ROOT_NAME,
                          *, trans_node: bool = None, num_workers: int = 0, **kwargs) -> None:
        assert not self._lazy_init.flag, 'Cannot add node group after document started'
        DocImpl._create_node_group_impl(self, 'node_groups', name=name, transform=transform, parent=parent,
                                        trans_node=trans_node, num_workers=num_workers, **kwargs)

    @classmethod
    def register_global_reader(cls, pattern: str, func: Optional[Callable] = None):
        if func is not None:
            cls._registered_file_reader[pattern] = func
            return None

        def decorator(klass):
            if callable(klass): cls._registered_file_reader[pattern] = klass
            else: raise TypeError(f"The registered object {klass} is not a callable object.")
            return klass
        return decorator

    def _resolve_index_placeholder(self, value):
        if isinstance(value, StorePlaceholder): return self.store
        elif isinstance(value, EmbedPlaceholder): return self.embed
        return value

    def register_index(self, index_type: str, index_cls: IndexBase, *args, **kwargs) -> None:
        if bool(self._lazy_init.flag):
            args = [self._resolve_index_placeholder(arg) for arg in args]
            kwargs = {k: self._resolve_index_placeholder(v) for k, v in kwargs.items()}
            self.store.register_index(index_type, index_cls(*args, **kwargs))
        else:
            self.index_pending_registrations.append((index_type, index_cls, args, kwargs))

    def add_reader(self, pattern: str, func: Optional[Callable] = None):
        assert callable(func), 'func for reader should be callable'
        self._local_file_reader[pattern] = func

    def worker(self):
        is_first_run = True
        while True:
            # Apply meta changes
            rows = self._dlm.fetch_docs_changed_meta(self._kb_group_name)
            for row in rows:
                new_meta_dict = json.loads(row[1]) if row[1] else {}
                self.store.update_doc_meta(row[0], new_meta_dict)

            # Step 1: do doc-parsing, highest priority
            docs = self._dlm.get_docs_need_reparse(group=self._kb_group_name)
            if docs:
                filepaths = [doc.path for doc in docs]
                ids = [doc.doc_id for doc in docs]
                metadatas = [json.loads(doc.meta) if doc.meta else None for doc in docs]
                # update status and need_reparse
                self._dlm.update_kb_group(cond_file_ids=ids, cond_group=self._kb_group_name,
                                          new_status=DocListManager.Status.working, new_need_reparse=False)
                self._delete_doc_from_store(filepaths)
                self._add_doc_to_store(input_files=filepaths, ids=ids, metadatas=metadatas)
                self._dlm.update_kb_group(cond_file_ids=ids, cond_group=self._kb_group_name,
                                          new_status=DocListManager.Status.success)

            # Step 2: After doc is deleted from related kb_group, delete doc from db
            if self._kb_group_name == DocListManager.DEFAULT_GROUP_NAME:
                self._dlm.delete_unreferenced_doc()

            # Step 3: do doc-deleting
            ids, files, metadatas = self._list_files(status=DocListManager.Status.deleting)
            if files:
                self._delete_doc_from_store(files)
                self._dlm.delete_files_from_kb_group(ids, self._kb_group_name)

            # Step 4: do doc-adding
            ids, files, metadatas = self._list_files(status=DocListManager.Status.waiting,
                                                     upload_status=DocListManager.Status.success)
            if files:
                self._dlm.update_kb_group(cond_file_ids=ids, cond_group=self._kb_group_name,
                                          new_status=DocListManager.Status.working)
                self._add_doc_to_store(input_files=files, ids=ids, metadatas=metadatas)
                # change working to success while leaving other status unchanged
                self._dlm.update_kb_group(cond_file_ids=ids, cond_group=self._kb_group_name,
                                          cond_status_list=[DocListManager.Status.working],
                                          new_status=DocListManager.Status.success)
            if is_first_run:
                self._init_monitor_event.set()
            is_first_run = False
            time.sleep(10)

    def _list_files(
            self, status: Union[str, List[str]] = DocListManager.Status.all,
            upload_status: Union[str, List[str]] = DocListManager.Status.all
    ) -> Tuple[List[str], List[str], List[Dict]]:
        if self._doc_files: return None, self._doc_files, None
        ids, paths, metadatas = [], [], []
        for row in self._dlm.list_kb_group_files(group=self._kb_group_name, status=status,
                                                 upload_status=upload_status, details=True):
            ids.append(row[0])
            paths.append(row[1])
            metadatas.append(json.loads(row[3]) if row[3] else {})
        return ids, paths, metadatas

    def _add_doc_to_store(self, input_files: List[str], ids: Optional[List[str]] = None,
                          metadatas: Optional[List[Dict[str, Any]]] = None):
        if not input_files:
            return
        root_nodes = self._reader.load_data(input_files)
        map_file_meta = {}
        if metadatas:
            for file_path, metadata in zip(input_files, metadatas):
                map_file_meta[file_path] = metadata
        for node in root_nodes:
            file_path = node.global_metadata[RAG_DOC_PATH]
            if file_path in map_file_meta:
                node.global_metadata.update(map_file_meta[file_path])
            node.global_metadata[RAG_DOC_ID] = gen_docid(node.docpath)
        temp_store = self._create_store({"type": "map"})
        temp_store.update_nodes(root_nodes)
        all_groups = self.store.all_groups()
        LOG.info(f"add_files: Trying to merge store with {all_groups}")
        for group in all_groups:
            if group != LAZY_ROOT_NAME and not self.store.is_group_active(group):
                continue
            # Duplicate group will be discarded automatically
            nodes = self._get_nodes(group, temp_store)
            self.store.update_nodes(nodes)
            LOG.debug(f"Merge {group} with {nodes}")

    def _delete_doc_from_store(self, input_files: List[str]) -> None:
        docs = self.store.get_index(type='file_node_map').query(input_files)
        LOG.info(f"delete_files: removing documents {input_files} and nodes {docs}")
        if len(docs) == 0:
            return
        self._delete_nodes_recursively(docs)

    def _delete_nodes_recursively(self, root_nodes: List[DocNode]) -> None:
        uids_to_delete = defaultdict(list)
        uids_to_delete[LAZY_ROOT_NAME] = [node._uid for node in root_nodes]

        # Gather all nodes to be deleted including their children
        def gather_children(node: DocNode):
            for children_group, children_list in node.children.items():
                for child in children_list:
                    uids_to_delete[children_group].append(child._uid)
                    gather_children(child)

        for node in root_nodes:
            gather_children(node)

        # Delete nodes in all groups
        for group, node_uids in uids_to_delete.items():
            self.store.remove_nodes(group, node_uids)
            LOG.debug(f"Removed nodes from group {group} for node IDs: {node_uids}")

    def _dynamic_create_nodes(self, group_name: str, store: StoreBase) -> None:
        if group_name == LAZY_ROOT_NAME or store.is_group_active(group_name):
            return
        node_group = self.node_groups.get(group_name)
        if node_group is None:
            raise ValueError(f"Node group '{group_name}' does not exist. Please check the group name "
                             "or add a new one through `create_node_group`.")
        t = node_group['transform']
        transform = AdaptiveTransform(t) if isinstance(t, list) or t.pattern else make_transform(t)
        parent_nodes = self._get_nodes(node_group["parent"], store)
        nodes = transform.batch_forward(parent_nodes, group_name)
        store.update_nodes(nodes)

    def _get_nodes(self, group_name: str, store: Optional[StoreBase] = None) -> List[DocNode]:
        store = store or self.store
        self._dynamic_create_nodes(group_name, store)
        return store.get_nodes(group_name)

    def retrieve(self, query: str, group_name: str, similarity: str, similarity_cut_off: Union[float, Dict[str, float]],
                 index: str, topk: int, similarity_kws: dict, embed_keys: Optional[List[str]] = None,
                 filters: Optional[Dict[str, Union[str, int, List, Set]]] = None) -> List[DocNode]:
        self._lazy_init()
        self._dynamic_create_nodes(group_name, self.store)

        if index is None or index == 'default':
            return self.store.query(query=query, group_name=group_name, similarity_name=similarity,
                                    similarity_cut_off=similarity_cut_off, topk=topk,
                                    embed_keys=embed_keys, filters=filters, **similarity_kws)

        index_instance = self.store.get_index(type=index)
        if not index_instance:
            raise NotImplementedError(f"index type '{index}' is not supported currently.")

        try:
            return index_instance.query(query=query, group_name=group_name, similarity_name=similarity,
                                        similarity_cut_off=similarity_cut_off, topk=topk,
                                        embed_keys=embed_keys, filters=filters, **similarity_kws)
        except Exception as e:
            raise RuntimeError(f'index type `{index}` of store `{type(self.store)}` query failed: {e}')

    def find(self, nodes: List[DocNode], group: str) -> List[DocNode]:
        if len(nodes) == 0: return nodes
        self._lazy_init()
        self._dynamic_create_nodes(group, self.store)

        def get_depth(name):
            cnt = 0
            while name != LAZY_ROOT_NAME:
                cnt += 1
                name = self.node_groups[name]['parent']
            return cnt

        # 1. find lowest common ancestor
        left, right = nodes[0]._group, group
        curr_depth, target_depth = get_depth(left), get_depth(right)
        if curr_depth > target_depth:
            for i in range(curr_depth - target_depth): left = self.node_groups[left]['parent']
        elif curr_depth < target_depth:
            for i in range(target_depth - curr_depth): right = self.node_groups[right]['parent']
        while (left != right):
            left = self.node_groups[left]['parent']
            right = self.node_groups[right]['parent']
        ancestor = left

        # 2. if ancestor != current group, go to ancestor; then if ancestor != target group, go to target group
        if nodes and nodes[0]._group != ancestor:
            nodes = DocImpl.find_parent(nodes, ancestor)
        if nodes and nodes[0]._group != group:
            nodes = DocImpl.find_children(nodes, group)
        return nodes

    @staticmethod
    def find_parent(nodes: List[DocNode], group: str) -> List[DocNode]:
        def recurse_parents(node: DocNode, visited: Set[DocNode]) -> None:
            if node.parent:
                if node.parent._group == group:
                    visited.add(node.parent)
                else:
                    recurse_parents(node.parent, visited)

        result = set()
        for node in nodes:
            recurse_parents(node, result)
        if not result:
            LOG.warning(
                f"We can not find any nodes for group `{group}`, please check your input"
            )
        LOG.debug(f"Found parent node for {group}: {result}")
        return list(result)

    @staticmethod
    def find_children(nodes: List[DocNode], group: str) -> List[DocNode]:
        def recurse_children(node: DocNode, visited: Set[DocNode]) -> bool:
            if group in node.children:
                visited.update(node.children[group])
                return True

            found_in_any_child = False

            for children_list in node.children.values():
                for child in children_list:
                    if recurse_children(child, visited):
                        found_in_any_child = True
                    else:
                        break

            return found_in_any_child

        result = set()

        for node in nodes:
            if group in node.children:
                result.update(node.children[group])
            else:
                if not recurse_children(node, result):
                    LOG.warning(
                        f"Node {node} and its children do not contain any nodes with the group `{group}`. "
                        "Skipping further search in this branch."
                    )
                    break

        if not result:
            LOG.warning(
                f"We cannot find any nodes for group `{group}`, please check your input."
            )

        LOG.debug(f"Found children nodes for {group}: {result}")
        return list(result)

    def clear_cache(self, group_names: Optional[List[str]] = None):
        self.store.clear_cache(group_names)

    def __call__(self, func_name: str, *args, **kwargs):
        return getattr(self, func_name)(*args, **kwargs)


for k, v in BuiltinGroups.__dict__.items():
    if not k.startswith('_') and isinstance(v, BuiltinGroups.Struct):
        assert k == v.name, 'builtin group name mismatch'
        DocImpl._create_builtin_node_group(name=k, transform=v.args, parent=v.parent)



import json
import os
import uuid
import requests
from typing import Tuple, List
from urllib.parse import urljoin
import lazyllm
from .onlineChatModuleBase import OnlineChatModuleBase
from .fileHandler import FileHandlerBase

class OpenAIModule(OnlineChatModuleBase, FileHandlerBase):
    TRAINABLE_MODEL_LIST = ["gpt-3.5-turbo-0125", "gpt-3.5-turbo-1106",
                            "gpt-3.5-turbo-0613", "babbage-002",
                            "davinci-002", "gpt-4-0613"]

    def __init__(self,
                 base_url: str = "https://api.openai.com/v1/",
                 model: str = "gpt-3.5-turbo",
                 api_key: str = None,
                 stream: bool = True,
                 return_trace: bool = False,
                 **kwargs):
        OnlineChatModuleBase.__init__(self,
                                      model_series="OPENAI",
                                      api_key=api_key or lazyllm.config['openai_api_key'],
                                      base_url=base_url,
                                      model_name=model,
                                      stream=stream,
                                      trainable_models=OpenAIModule.TRAINABLE_MODEL_LIST,
                                      return_trace=return_trace,
                                      **kwargs)
        FileHandlerBase.__init__(self)
        self.default_train_data = {
            "model": "gpt-3.5-turbo-0613",
            "training_file": None,
            "validation_file": None,
            "hyperparameters": {
                "n_epochs": 1,
                "batch_size": 16,
                "learning_rate_multiplier": "1.6e-5",
            }
        }
        self.fine_tuning_job_id = None

    def _get_system_prompt(self):
        return "You are ChatGPT, a large language model trained by OpenAI.You are a helpful assistant."

    def _convert_file_format(self, filepath: str) -> str:
        with open(filepath, 'r', encoding='utf-8') as fr:
            dataset = [json.loads(line) for line in fr]

        json_strs = []
        for ex in dataset:
            lineEx = {"messages": []}
            messages = ex.get("messages", [])
            for message in messages:
                role = message.get("role", "")
                content = message.get("content", "")
                if role in ["system", "user", "assistant"]:
                    lineEx["messages"].append({"role": role, "content": content})
            json_strs.append(json.dumps(lineEx, ensure_ascii=False))

        return "\n".join(json_strs)

    def _upload_train_file(self, train_file):
        headers = {
            "Authorization": "Bearer " + self._api_key
        }

        url = urljoin(self._base_url, "files")

        self.get_finetune_data(train_file)

        file_object = {
            "purpose": (None, "fine-tune", None),
            "file": (os.path.basename(train_file), self._dataHandler, "application/json")
        }

        with requests.post(url, headers=headers, files=file_object) as r:
            if r.status_code != 200:
                raise requests.RequestException('\n'.join([c.decode('utf-8') for c in r.iter_content(None)]))

            # delete temporary training file
            self._dataHandler.close()
            return r.json()["id"]

    def _update_kw(self, data, normal_config):
        current_train_data = self.default_train_data.copy()
        current_train_data.update(data)

        current_train_data["hyperparameters"]["n_epochs"] = normal_config["num_epochs"]
        current_train_data["hyperparameters"]["learning_rate_multiplier"] = str(normal_config["learning_rate"])
        current_train_data["hyperparameters"]["batch_size"] = normal_config["batch_size"]
        current_train_data["suffix"] = str(uuid.uuid4())[:7]

        return current_train_data

    def _create_finetuning_job(self, train_model, train_file_id, **kw) -> Tuple[str, str]:
        url = urljoin(self._base_url, "fine_tuning/jobs")
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._api_key}",
        }
        data = {
            "model": train_model,
            "training_file": train_file_id
        }
        if len(kw) > 0:
            if 'finetuning_type' in kw:
                data = self._update_kw(data, kw)
            else:
                data.update(kw)

        with requests.post(url, headers=headers, json=data) as r:
            if r.status_code != 200:
                raise requests.RequestException('\n'.join([c.decode('utf-8') for c in r.iter_content(None)]))

            fine_tuning_job_id = r.json()["id"]
            self.fine_tuning_job_id = fine_tuning_job_id
            status = r.json()["status"]
            return (fine_tuning_job_id, status)

    def _cancel_finetuning_job(self, fine_tuning_job_id=None):
        if not fine_tuning_job_id and not self.fine_tuning_job_id:
            return 'Invalid'
        job_id = fine_tuning_job_id if fine_tuning_job_id else self.fine_tuning_job_id
        fine_tune_url = urljoin(self._base_url, f"fine_tuning/jobs/{job_id}/cancel")
        headers = {
            "Authorization": f"Bearer {self._api_key}"
        }
        with requests.post(fine_tune_url, headers=headers) as r:
            if r.status_code != 200:
                raise requests.RequestException('\n'.join([c.decode('utf-8') for c in r.iter_content(None)]))
        status = r.json()['status']
        if status == 'cancelled':
            return 'Cancelled'
        else:
            return f'JOB {job_id} status: {status}'

    def _query_finetuned_jobs(self):
        fine_tune_url = urljoin(self._base_url, "fine_tuning/jobs")
        headers = {
            "Authorization": f"Bearer {self._api_key}",
        }
        with requests.get(fine_tune_url, headers=headers) as r:
            if r.status_code != 200:
                raise requests.RequestException('\n'.join([c.decode('utf-8') for c in r.iter_content(None)]))
        return r.json()

    def _get_finetuned_model_names(self) -> Tuple[List[Tuple[str, str]], List[Tuple[str, str]]]:
        model_data = self._query_finetuned_jobs()
        res = list()
        for model in model_data['data']:
            res.append([model['id'], model['fine_tuned_model'], self._status_mapping(model['status'])])
        return res

    def _status_mapping(self, status):
        if status == 'succeeded':
            return 'Done'
        elif status == 'failed':
            return 'Failed'
        elif status == 'cancelled':
            return 'Cancelled'
        elif status == 'running':
            return 'Running'
        else:  # validating_files, queued
            return 'Pending'

    def _query_job_status(self, fine_tuning_job_id=None):
        if not fine_tuning_job_id and not self.fine_tuning_job_id:
            raise RuntimeError("No job ID specified. Please ensure that a valid 'fine_tuning_job_id' is "
                               "provided as an argument or started a training job.")
        job_id = fine_tuning_job_id if fine_tuning_job_id else self.fine_tuning_job_id
        _, status = self._query_finetuning_job(job_id)
        return self._status_mapping(status)

    def _get_log(self, fine_tuning_job_id=None):
        if not fine_tuning_job_id and not self.fine_tuning_job_id:
            raise RuntimeError("No job ID specified. Please ensure that a valid 'fine_tuning_job_id' is "
                               "provided as an argument or started a training job.")
        job_id = fine_tuning_job_id if fine_tuning_job_id else self.fine_tuning_job_id
        fine_tune_url = urljoin(self._base_url, f"fine_tuning/jobs/{job_id}/events")
        headers = {
            "Authorization": f"Bearer {self._api_key}"
        }
        with requests.get(fine_tune_url, headers=headers) as r:
            if r.status_code != 200:
                raise requests.RequestException('\n'.join([c.decode('utf-8') for c in r.iter_content(None)]))
        return job_id, r.json()

    def _get_curr_job_model_id(self):
        if not self.fine_tuning_job_id:
            return None, None
        model_id, _ = self._query_finetuning_job(self.fine_tuning_job_id)
        return self.fine_tuning_job_id, model_id

    def _query_finetuning_job_info(self, fine_tuning_job_id):
        fine_tune_url = urljoin(self._base_url, f"fine_tuning/jobs/{fine_tuning_job_id}")
        headers = {
            "Authorization": f"Bearer {self._api_key}"
        }
        with requests.get(fine_tune_url, headers=headers) as r:
            if r.status_code != 200:
                raise requests.RequestException('\n'.join([c.decode('utf-8') for c in r.iter_content(None)]))
        return r.json()

    def _query_finetuning_job(self, fine_tuning_job_id) -> Tuple[str, str]:
        info = self._query_finetuning_job_info(fine_tuning_job_id)
        status = info['status']
        fine_tuned_model = info["fine_tuned_model"] if 'fine_tuned_model' in info else None
        return (fine_tuned_model, status)

    def _query_finetuning_cost(self, fine_tuning_job_id):
        info = self._query_finetuning_job_info(fine_tuning_job_id)
        if 'trained_tokens' in info and info['trained_tokens']:
            return info['trained_tokens']
        else:
            return None

    def _create_deployment(self) -> Tuple[str, str]:
        return (self._model_name, "RUNNING")

    def _query_deployment(self, deployment_id) -> str:
        return "RUNNING"



import os
import socket
import requests
import json
from typing import Union

import lazyllm
from lazyllm import LOG
from lazyllm import ModuleBase, ServerModule
from lazyllm.thirdparty import gradio as gr
from lazyllm.flow import Pipeline


class WebUi:
    def __init__(self, base_url) -> None:
        self.base_url = base_url

    def basic_headers(self, content_type=True):
        return {
            "accept": "application/json",
            "Content-Type": "application/json" if content_type else None,
        }

    def muti_headers(
        self,
    ):
        return {"accept": "application/json"}

    def post_request(self, url, data):
        response = requests.post(
            url, headers=self.basic_headers(), data=json.dumps(data)
        )
        return response.json()

    def get_request(self, url):
        response = requests.get(url, headers=self.basic_headers(False))
        return response.json()

    def new_group(self, group_name: str):
        response = requests.post(
            f"{self.base_url}/new_group?group_name={group_name}",
            headers=self.basic_headers(True),
        )
        return response.json()["msg"]

    def delete_group(self, group_name: str):
        response = requests.post(
            f"{self.base_url}/delete_group?group_name={group_name}",
            headers=self.basic_headers(True),
        )
        return response.json()["msg"]

    def list_groups(self):
        response = requests.get(
            f"{self.base_url}/list_kb_groups", headers=self.basic_headers(False)
        )
        return response.json()["data"]

    def upload_files(self, group_name: str, override: bool = True):
        response = requests.post(
            f"{self.base_url}/upload_files?group_name={group_name}&override={override}",
            headers=self.basic_headers(True),
        )
        return response.json()["data"]

    def list_files_in_group(self, group_name: str):
        response = requests.get(
            f"{self.base_url}/list_files_in_group?group_name={group_name}&alive=True",
            headers=self.basic_headers(False),
        )
        return response.json()["data"]

    def delete_file(self, group_name: str, file_ids: list[str]):
        response = requests.post(
            f"{self.base_url}/delete_files_from_group",
            headers=self.basic_headers(True),
            json={"group_name": group_name, "file_ids": file_ids}
        )
        return response.json()["msg"]

    def gr_show_list(self, str_list: list, list_name: Union[str, list]):
        if isinstance(list_name, str):
            headers = ["index", list_name]
            value = [[index, str_list[index]] for index in range(len(str_list))]
        else:
            headers = ["index"] + list_name
            value = [[index] + str_list[index:index + len(list_name)] for index in range(len(str_list))]
        return gr.DataFrame(headers=headers, value=value)

    def create_ui(self):
        with gr.Blocks(analytics_enabled=False) as demo:
            with gr.Tabs():
                select_group_list = []

                with gr.TabItem("分组列表"):
                    select_group = self.gr_show_list(
                        self.list_groups(), list_name="group_name"
                    )
                    select_group_list.append(select_group)

                with gr.TabItem("上传文件"):

                    def _upload_files(group_name, files):

                        files_to_upload = [
                            ("files", (os.path.basename(file), open(file, "rb")))
                            for file in files
                        ]

                        url = f"{self.base_url}/add_files_to_group?group_name={group_name}&override=true"
                        response = requests.post(
                            url, files=files_to_upload, headers=self.muti_headers()
                        )
                        response.raise_for_status()
                        response_data = response.json()
                        gr.Info(str(response_data["msg"]))

                        for _, (_, file_obj) in files_to_upload:
                            file_obj.close()

                    select_group = gr.Dropdown(self.list_groups(), label="选择分组")
                    select_group.change(lambda x: x, inputs=select_group, outputs=None)

                    up_files = gr.Files(label="上传文件")
                    up_btn = gr.Button("上传")
                    up_btn.click(
                        _upload_files,
                        inputs=[select_group, up_files],
                        outputs=None,
                    )

                    select_group_list.append(select_group)

                with gr.TabItem("分组文件列表"):
                    def _list_group_files(group_name):
                        file_list = self.list_files_in_group(group_name)
                        values = [[i] + file_list[i][:2] for i in range(len(file_list))]
                        return gr.update(
                            value=values
                        )

                    select_group = gr.Dropdown(self.list_groups(), label="选择分组")
                    show_list = self.gr_show_list([], list_name=["file_id", "file_name"])
                    select_group.change(
                        fn=_list_group_files, inputs=select_group, outputs=show_list
                    )
                    select_group_list.append(select_group)

                with gr.TabItem("删除文件"):

                    def _list_group_files(group_name):
                        file_list = self.list_files_in_group(group_name)
                        file_list = [','.join(file[:2]) for file in file_list]
                        return gr.update(choices=file_list)

                    select_group = gr.Dropdown(self.list_groups(), label="选择分组")
                    select_file = gr.Dropdown([], label="选择文件")
                    select_group.change(
                        fn=_list_group_files, inputs=select_group, outputs=select_file
                    )
                    delete_btn = gr.Button("删除")

                    def _delete_file(group_name, select_file):
                        file_ids = [select_file.split(',')[0]]
                        gr.Info(self.delete_file(group_name, file_ids))
                        return _list_group_files(group_name)

                    delete_btn.click(
                        fn=_delete_file,
                        inputs=[select_group, select_file],
                        outputs=select_file,
                    )
                    select_group_list.append(select_group)

        return demo


class DocWebModule(ModuleBase):
    class Mode:
        Dynamic = 0
        Refresh = 1
        Appendix = 2

    def __init__(
        self,
        doc_server: ServerModule,
        title="文档管理演示终端",
        port=range(20800, 20999),
        history=[],
        text_mode=None,
        trace_mode=None,
    ) -> None:
        super().__init__()
        self.title = title
        self.port = port
        self.history = history
        self.trace_mode = trace_mode if trace_mode else DocWebModule.Mode.Refresh
        self.text_mode = text_mode if text_mode else DocWebModule.Mode.Dynamic
        self.doc_server = doc_server
        self._deploy_flag = lazyllm.once_flag()
        self.api_url = ""
        self.url = ""

    def _prepare(self, query, chat_history):
        if chat_history is None:
            chat_history = []
        return "", chat_history + [[query, None]]

    def _clear_history(self):
        return [], "", ""

    def _work(self):
        if isinstance(self.port, (range, tuple, list)):
            port = self._find_can_use_network_port()
        else:
            port = self.port
            assert self._verify_port_access(port), f"port {port} is occupied"

        self.api_url = self.doc_server._url.rsplit("/", 1)[0]
        self.web_ui = WebUi(self.api_url)
        self.demo = self.web_ui.create_ui()
        self.url = f'http://127.0.0.1:{port}'
        self.broadcast_url = f'http://0.0.0.0:{port}'

        self.demo.queue().launch(server_name="0.0.0.0", server_port=port, prevent_thread_lock=True)
        LOG.success('LazyLLM docwebmodule launched successfully: Running on: '
                    f'{self.broadcast_url}, local URL: {self.url}', flush=True)

    def _get_deploy_tasks(self):
        return Pipeline(self._work)

    def _get_post_process_tasks(self):
        return Pipeline(self._print_url)

    def wait(self):
        self.demo.block_thread()

    def stop(self):
        if self.demo:
            self.demo.close()
            del self.demo
            self.demo, self.url = None, ''

    def _find_can_use_network_port(self):
        for port in self.port:
            if self._verify_port_access(port):
                return port
        raise RuntimeError(
            f"The ports in the range {self.port} are all occupied. "
            "Please change the port range or release the relevant ports."
        )

    def _print_url(self):
        lazyllm.LOG.success(
            f"LazyLLM DocWebModule launched successfully: Running on local URL: {self.url}",
            flush=True,
        )

    def _verify_port_access(self, port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            result = s.connect_ex(("127.0.0.1", port))
            return result != 0

    def __repr__(self):
        return lazyllm.make_repr("Module", "DocWebModule")



from urllib.parse import urljoin
import lazyllm
from .onlineChatModuleBase import OnlineChatModuleBase


class DeepSeekModule(OnlineChatModuleBase):
    def __init__(self,
                 base_url: str = "https://api.deepseek.com",
                 model: str = "deepseek-chat",
                 api_key: str = None,
                 stream: bool = True,
                 return_trace: bool = False,
                 **kwargs):
        super().__init__(model_series="DEEPSEEK",
                         api_key=api_key or lazyllm.config['deepseek_api_key'],
                         base_url=base_url,
                         model_name=model,
                         stream=stream,
                         trainable_models=[],
                         return_trace=return_trace,
                         **kwargs)

    def _get_system_prompt(self):
        return "You are an intelligent assistant developed by China's DeepSeek. You are a helpful assistanti."

    def _set_chat_url(self):
        self._url = urljoin(self._base_url, "chat/completions")



from typing import List, Optional
from .doc_node import DocNode
from .index_base import IndexBase
from lazyllm.common import override
from .map_store import MapStore
from .milvus_store import MilvusStore

class SmartEmbeddingIndex(IndexBase):
    def __init__(self, backend_type: str, **kwargs):
        if backend_type == 'milvus':
            self._store = MilvusStore(**kwargs)
        elif backend_type == 'map':
            self._store = MapStore(**kwargs)
        else:
            raise ValueError(f'unsupported backend [{backend_type}]')

    @override
    def update(self, nodes: List[DocNode]) -> None:
        self._store.update_nodes(nodes)

    @override
    def remove(self, uids: List[str], group_name: Optional[str] = None) -> None:
        self._store.remove_nodes(group_name, uids)

    @override
    def query(self, *args, **kwargs) -> List[DocNode]:
        return self._store.query(*args, **kwargs)



import re
from pathlib import Path
from typing import Dict, List, Optional
from fsspec import AbstractFileSystem

from .readerBase import LazyLLMReaderBase
from ..doc_node import DocNode

class IPYNBReader(LazyLLMReaderBase):
    def __init__(self, parser_config: Optional[Dict] = None, concatenate: bool = False, return_trace: bool = True):
        super().__init__(return_trace=return_trace)
        self._parser_config = parser_config
        self._concatenate = concatenate

    def _load_data(self, file: Path, extra_info: Optional[Dict] = None,
                   fs: Optional[AbstractFileSystem] = None) -> List[DocNode]:
        if not isinstance(file, Path): file = Path(file)

        if file.name.endswith(".ipynb"):
            try:
                import nbconvert
            except ImportError:
                raise ImportError("Please install nbconvert `pip install nbconvert`")

        if fs:
            with fs.open(file, encoding='utf-8') as f:
                doc_str = nbconvert.exporters.ScriptExporter().from_file(f)[0]
        else:
            doc_str = nbconvert.exporters.ScriptExporter().from_file(file)[0]

        splits = re.split(r"In\[\d+\]:", doc_str)
        splits.pop(0)

        if self._concatenate: docs = [DocNode(text="\n\n".join(splits), metadata=extra_info or {})]
        else: docs = [DocNode(text=s, global_metadata=extra_info) for s in splits]

        return docs
